# -*- coding: utf-8 -*-
# File: hf_detr_train.py

# Copyright 2023 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for training Hugging Face Detr implementation. Note, that this scripts only trans Tabletransformer like Detr
models that are a slightly different from the plain Detr model that are provided by the transformer library.
"""

import copy
from typing import Any, Dict, List, Optional, Sequence, Type, Union

from torch.nn import Module
from torch.utils.data import Dataset
from transformers import (
    AutoFeatureExtractor,
    IntervalStrategy,
    PretrainedConfig,
    PreTrainedModel,
    TableTransformerForObjectDetection,
)
from transformers.trainer import Trainer, TrainingArguments

from ..datasets.adapter import DatasetAdapter
from ..datasets.base import DatasetBase
from ..datasets.registry import get_dataset
from ..eval.base import MetricBase
from ..eval.eval import Evaluator
from ..eval.registry import metric_registry
from ..extern.hfdetr import HFDetrDerivedDetector
from ..mapper.hfstruct import DetrDataCollator, image_to_hf_detr_training
from ..pipe.base import PredictorPipelineComponent
from ..pipe.registry import pipeline_component_registry
from ..utils.logger import logger
from ..utils.utils import string_to_dict


class DetrDerivedTrainer(Trainer):
    """
    Huggingface Trainer for training Transformer models with a custom evaluate method in order
    to use dd Evaluator. Train setting is not defined in the trainer itself but in config setting as
    defined in `TrainingArguments`. Please check the Transformer documentation

    <https://huggingface.co/docs/transformers/main_classes/trainer>

    for custom training setting.
    """

    def __init__(
        self,
        model: Union[PreTrainedModel, Module],
        args: TrainingArguments,
        data_collator: DetrDataCollator,
        train_dataset: Dataset[Any],
    ):
        self.evaluator: Optional[Evaluator] = None
        self.build_eval_kwargs: Optional[Dict[str, Any]] = None
        super().__init__(model, args, data_collator, train_dataset)

    def setup_evaluator(
        self,
        dataset_val: DatasetBase,
        pipeline_component: PredictorPipelineComponent,
        metric: Union[Type[MetricBase], MetricBase],
        **build_eval_kwargs: Union[str, int],
    ) -> None:
        """
        Setup of evaluator before starting training. During training, predictors will be replaced by current
        checkpoints.

        :param dataset_val: dataset on which to run evaluation
        :param pipeline_component: pipeline component to plug into the evaluator
        :param metric: A metric class
        :param build_eval_kwargs:
        """

        self.evaluator = Evaluator(dataset_val, pipeline_component, metric, num_threads=1)
        assert self.evaluator.pipe_component
        for comp in self.evaluator.pipe_component.pipe_components:
            assert isinstance(comp, PredictorPipelineComponent)
            assert isinstance(comp.predictor, HFDetrDerivedDetector)
            comp.predictor.hf_detr_predictor = None
        self.build_eval_kwargs = build_eval_kwargs

    def evaluate(
        self,
        eval_dataset: Optional[Dataset[Any]] = None,
        ignore_keys: Optional[List[str]] = None,
        metric_key_prefix: str = "eval",
    ) -> Dict[str, float]:
        """
        Overwritten method from `Trainer`. Arguments will not be used.
        """
        assert self.evaluator is not None
        assert self.evaluator.pipe_component is not None

        # memory metrics - must set up as early as possible
        self._memory_tracker.start()
        for comp in self.evaluator.pipe_component.pipe_components:
            comp.predictor.hf_detr_predictor = copy.deepcopy(self.model).eval()  # type: ignore
        if isinstance(self.build_eval_kwargs, dict):
            scores = self.evaluator.run(True, **self.build_eval_kwargs)
        else:
            scores = self.evaluator.run(True)

        self.log(scores)

        return scores


def train_hf_detr(
    path_config_json: str,
    dataset_train: Union[str, DatasetBase],
    path_weights: str,
    path_feature_extractor_config_json: str,
    config_overwrite: Optional[List[str]] = None,
    log_dir: str = "train_log/detr",
    build_train_config: Optional[Sequence[str]] = None,
    dataset_val: Optional[DatasetBase] = None,
    build_val_config: Optional[Sequence[str]] = None,
    metric_name: Optional[str] = None,
    metric: Optional[Union[Type[MetricBase], MetricBase]] = None,
    pipeline_component_name: Optional[str] = None,
) -> None:
    """
    Train Tabletransformer from scratch or fine-tune using an adaptation of the transformer trainer.
    Allowing experiments by using different config settings.

    :param path_config_json: path to a Tabletransformer config file
    :param dataset_train: dataset to use for training
    :param path_weights: path to a checkpoint, if you want to resume training or fine-tune. Will train from scratch if
                         an empty string is passed
    :param path_feature_extractor_config_json: path to a feature extractor config file. In many situations you can use
                                               the standard config file:

                                                   ModelCatalog.
                                                   get_full_path_preprocessor_configs
                                                   ("microsoft/table-transformer-detection/pytorch_model.bin")

    :param config_overwrite: Pass a list of arguments if some configs from the .json file are supposed to be replaced.
                             Use the list convention, e.g. ['per_device_train_batch_size=4']
    :param log_dir: Will default to 'train_log/detr'
    :param build_train_config: dataflow build setting. Again, use list convention setting, e.g. ['max_datapoints=1000']
    :param dataset_val: the dataset to use for validation
    :param build_val_config: same as `build_train_config` but for dataflow validation
    :param metric_name: A metric name to choose for validation. Will use the default setting. If you want a custom
                        metric setting, pass a metric explicitly.
    :param metric: A metric to choose for validation
    :param pipeline_component_name: A pipeline component name to use for validation
    """

    build_train_dict: Dict[str, str] = {}
    if build_train_config is not None:
        build_train_dict = string_to_dict(",".join(build_train_config))
    if "split" not in build_train_dict:
        build_train_dict["split"] = "train"

    build_val_dict: Dict[str, str] = {}
    if build_val_config is not None:
        build_val_dict = string_to_dict(",".join(build_val_config))
    if "split" not in build_val_dict:
        build_val_dict["split"] = "val"

    if config_overwrite is None:
        config_overwrite = []

    if isinstance(dataset_train, str):
        dataset_train = get_dataset(dataset_train)

    categories = dataset_train.dataflow.categories.get_categories(as_dict=False, filtered=True)
    categories_dict_name_as_key = dataset_train.dataflow.categories.get_categories(name_as_key=True, filtered=True)

    dataset = DatasetAdapter(
        dataset_train, True, image_to_hf_detr_training(category_names=categories), True, **build_train_dict
    )

    number_samples = len(dataset)
    conf_dict = {
        "output_dir": log_dir,
        "remove_unused_columns": False,
        "per_device_train_batch_size": 2,
        "max_steps": number_samples,
        "evaluation_strategy": "steps"
        if (dataset_val is not None and metric is not None and pipeline_component_name is not None)
        else "no",
        "eval_steps": 5000,
    }

    for conf in config_overwrite:
        key, val = conf.split("=", maxsplit=1)
        try:
            val = int(val)  # type: ignore
        except ValueError:
            try:
                val = float(val)  # type: ignore
            except ValueError:
                pass
        conf_dict[key] = val

    # Will inform about dataloader warnings if max_steps exceeds length of dataset
    if conf_dict["max_steps"] > number_samples:  # type: ignore
        logger.warning(
            "After %s dataloader will log warning at every iteration about unexpected samples", number_samples
        )

    arguments = TrainingArguments(**conf_dict)
    logger.info("Config: \n %s", str(arguments.to_dict()), arguments.to_dict())

    id2label = {int(k) - 1: v for v, k in categories_dict_name_as_key.items()}
    config = PretrainedConfig.from_pretrained(
        pretrained_model_name_or_path=path_config_json,
        num_labels=len(id2label),
    )

    if path_weights != "":
        model = TableTransformerForObjectDetection.from_pretrained(
            pretrained_model_name_or_path=path_weights, config=config
        )
    else:
        model = TableTransformerForObjectDetection(config)

    feature_extractor = AutoFeatureExtractor.from_pretrained(
        pretrained_model_name_or_path=path_feature_extractor_config_json
    )
    data_collator = DetrDataCollator(feature_extractor)
    trainer = DetrDerivedTrainer(model, arguments, data_collator, dataset)

    if arguments.evaluation_strategy in (IntervalStrategy.STEPS,):
        categories = dataset_val.dataflow.categories.get_categories(filtered=True)  # type: ignore
        detector = HFDetrDerivedDetector(
            path_config_json, path_weights, path_feature_extractor_config_json, categories  # type: ignore
        )
        pipeline_component_cls = pipeline_component_registry.get(pipeline_component_name)
        pipeline_component = pipeline_component_cls(detector)
        assert isinstance(pipeline_component, PredictorPipelineComponent)

        if metric_name is not None:
            metric = metric_registry.get(metric_name)
        assert metric is not None

        trainer.setup_evaluator(dataset_val, pipeline_component, metric)  # type: ignore

    trainer.train()



# -*- coding: utf-8 -*-
# File: serialize.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
Some DataFlow classes for serialization. Many classes have been taken from

<https://github.com/tensorpack/dataflow/blob/master/dataflow/dataflow/raw.py>
"""

import pickle
from copy import copy
from typing import Any, Iterable, Iterator, List, Optional, Tuple, Union

import numpy as np

from .base import DataFlow, DataFlowResetStateNotCalled, RNGDataFlow


class DataFromList(RNGDataFlow):
    """Wrap a list of datapoints to a DataFlow"""

    def __init__(self, lst: List[Any], shuffle: bool = True) -> None:
        """
        :param lst: input list. Each element is a datapoint.
        :param shuffle: shuffle data.
        """
        super().__init__()
        self.lst = lst
        self.shuffle = shuffle

    def __len__(self) -> int:
        return len(self.lst)

    def __iter__(self) -> Iterator[Any]:
        if not self.shuffle:
            yield from self.lst
        else:
            idxs = np.arange(len(self.lst))
            if self.rng is not None:
                self.rng.shuffle(idxs)
                for k in idxs:
                    yield self.lst[k]
            else:
                raise DataFlowResetStateNotCalled()


class DataFromIterable(DataFlow):
    """Wrap an iterable of datapoints to a DataFlow"""

    def __init__(self, iterable: Iterable[Any]) -> None:
        """
        :param iterable: an iterable object
        """
        self._itr = iterable
        self._len: Optional[int] = None
        try:
            self._len = len(iterable)  # type: ignore
        except (NotImplementedError, TypeError):
            pass

    def __len__(self) -> int:
        if self._len is None:
            raise NotImplementedError
        return self._len

    def __iter__(self) -> Iterator[Any]:
        yield from self._itr

    def reset_state(self) -> None:
        pass


class FakeData(RNGDataFlow):
    """Generate fake data of given shapes"""

    def __init__(
        self,
        shapes: List[Union[List[Any], Tuple[Any]]],
        size: int = 1000,
        random: bool = True,
        dtype: str = "float32",
        domain: Tuple[Union[float, int], Union[float, int]] = (0, 1),
    ):
        """
        :param  shapes: a list of lists/tuples. Shapes of each component.
        :param  size: size of this DataFlow.
        :param  random: whether to randomly generate data every iteration.
                        Note that merely generating the data could sometimes be time-consuming!
        :param dtype: data type as string, or a list of data types.
        :param domain: (min, max) tuple, or a list of such tuples
        """

        super().__init__()
        self.shapes = shapes
        self._size = int(size)
        self.random = random
        self.dtype = [dtype] * len(shapes) if isinstance(dtype, str) else dtype
        self.domain = [domain] * len(shapes) if isinstance(domain, tuple) else domain
        if len(self.dtype) != len(self.shapes):
            raise ValueError(f"self.dtype={self.dtype} and self.shapes={self.shapes} must have same length")

    def __len__(self) -> int:
        return self._size

    def __iter__(self) -> Iterator[Any]:
        if self.rng is None:
            raise DataFlowResetStateNotCalled()
        if self.random:
            for _ in range(self._size):
                val = []
                for idx, _ in enumerate(self.shapes):
                    var = (
                        self.rng.rand(*self.shapes[idx]) * (self.domain[idx][1] - self.domain[idx][0])
                        + self.domain[idx][0]
                    )
                    val.append(var.astype(self.dtype[idx]))
                yield val
        else:
            val = []
            for idx, _ in enumerate(self.shapes):
                var = (
                    self.rng.rand(*self.shapes[idx]) * (self.domain[idx][1] - self.domain[idx][0]) + self.domain[idx][0]
                )
                val.append(var.astype(self.dtype[idx]))
            for _ in range(self._size):
                yield copy(val)


class PickleSerializer:
    """A Serializer to load and to dump objects"""

    @staticmethod
    def dumps(obj: Any) -> bytes:
        """
        :param obj: bytes
        """
        return pickle.dumps(obj, protocol=-1)

    @staticmethod
    def loads(buf: Any) -> Any:
        """
        :param buf: bytes
        """
        return pickle.loads(buf)



# -*- coding: utf-8 -*-
# File: language.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for language detection pipeline component
"""
from copy import copy, deepcopy
from typing import Optional, Sequence

from ..datapoint.image import Image
from ..datapoint.view import Page
from ..extern.base import LanguageDetector, ObjectDetector
from ..utils.detection_types import JsonDict
from ..utils.logger import logger
from ..utils.settings import LayoutType, PageType, TypeOrStr, get_type
from .base import PipelineComponent
from .registry import pipeline_component_registry


@pipeline_component_registry.register("LanguageDetectionService")
class LanguageDetectionService(PipelineComponent):
    """
    Pipeline Component for identifying the language in an image.

    There are two ways to use this component:

    - By analyzing the already extracted and ordered text. For this purpose, a page object is parsed internally and
      the full text is passed to the language_detector. This approach provides the greatest precision.

    - By previous text extraction with an object detector and subsequent transfer of concatenated word elements to the
      language_detector. Only one OCR detector can be used here. This method can be used, for example, to select an OCR
      detector that specializes in a language using. Although the word recognition is less accurate
      when choosing any detector, the results are confident enough to rely on the results, especially when extracting
      longer text passages. So, a TextExtractionService, for example, can be selected as the subsequent pipeline
      component. The words determined by the OCR detector are not transferred to the image object.

          lang_detector = FasttextLangDetector(path_weights,profile.categories)
          component = LanguageDetectionService(lang_detector, text_container="word",
                                               text_block_names=["text","title","table"])

    """

    def __init__(
        self,
        language_detector: LanguageDetector,
        text_container: Optional[TypeOrStr] = None,
        text_detector: Optional[ObjectDetector] = None,
        text_block_names: Optional[Sequence[TypeOrStr]] = None,
    ):
        """
        :param language_detector: Detector to determine text
        :param text_container: text container, needed to generate the reading order. Not necessary when passing a
                               text detector.
        :param text_detector: Object detector to extract text. You cannot use a Pdfminer here.

        :param text_block_names: text blocks, needed for generating the reading order. Not necessary
                                 when passing a text detector.
        """

        self.predictor = language_detector
        self.text_detector = text_detector
        self.text_container = get_type(text_container) if text_container is not None else text_container
        _text_block_names = []
        if text_block_names:
            _text_block_names = [get_type(text_block) for text_block in text_block_names]
        self.text_block_names = _text_block_names
        self._init_sanity_checks()
        super().__init__(
            self._get_name(self.predictor.name)
        )  # cannot use PredictorPipelineComponent class because of return type of predict meth

    def serve(self, dp: Image) -> None:
        if self.text_detector is None:
            page = Page.from_image(dp, self.text_container, self.text_block_names)  # type: ignore
            text = page.text_no_line_break
        else:
            if dp.image is None:
                raise ValueError("dp.image cannot be None")
            detect_result_list = self.text_detector.predict(dp.image)
            # this is a concatenation of all detection result. No reading order
            text = " ".join([result.text for result in detect_result_list if result.text is not None])
        predict_result = self.predictor.predict(text)
        self.dp_manager.set_summary_annotation(
            PageType.language, PageType.language, 1, predict_result.text, predict_result.score
        )

    def _init_sanity_checks(self) -> None:
        assert (
            self.text_detector or self.text_container
        ), "if no text_detector is provided a text container must be specified"
        if not self.text_detector:
            assert self.text_container in [LayoutType.word, LayoutType.line], (
                f"text_container must be either {LayoutType.word} or " f"{LayoutType.line}"
            )
            if not self.text_block_names:
                logger.info("text_block_names are set to None. This setting will return no reading order!")

    def clone(self) -> PipelineComponent:
        predictor = self.predictor.clone()
        if not isinstance(predictor, LanguageDetector):
            raise ValueError(f"Predictor must be of type LanguageDetector, but is of type {type(predictor)}")
        return self.__class__(
            predictor,
            copy(self.text_container),
            deepcopy(self.text_detector),
            deepcopy(self.text_block_names),
        )

    def get_meta_annotation(self) -> JsonDict:
        return dict(
            [
                ("image_annotations", []),
                ("sub_categories", {}),
                ("relationships", {}),
                ("summaries", [PageType.language]),
            ]
        )

    @staticmethod
    def _get_name(predictor_name: str) -> str:
        return f"language_detection_{predictor_name}"



# -*- coding: utf-8 -*-
# File: test_info.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module datasets.info
"""

from typing import Tuple

import pytest

from deepdoctection.datasets import DatasetCategories
from deepdoctection.datasets.info import get_merged_categories
from deepdoctection.utils.settings import get_type

from ..data import TestType


class TestDatasetCategories:
    """
    Testing DatasetCategories
    """

    @staticmethod
    def setup() -> DatasetCategories:
        """
        Arrange testing setup
        """
        categories = [get_type("FOO"), get_type("BAK"), get_type("BAZ")]
        sub_categories = {
            get_type("BAK"): {
                get_type("sub"): [get_type("BAK_11"), get_type("BAK_12")],
                get_type("sub_2"): [get_type("BAK_21"), get_type("BAK_22")],
            },
            get_type("FOO"): {get_type("cat"): [get_type("FOO_1"), get_type("FOO_2"), get_type("FOO_3")]},
        }

        return DatasetCategories(init_categories=categories, init_sub_categories=sub_categories)

    @staticmethod
    @pytest.mark.basic
    def test_set_cat_to_subcat_and_check_categories_case_1() -> None:
        """
        Categories are dumped and annotation ids are correctly assigned and  is_cat_to_sub_cat works properly.
        """

        # Arrange
        cats = TestDatasetCategories.setup()

        # Assert
        assert cats.get_categories() == {"1": TestType.FOO, "2": TestType.BAK, "3": TestType.BAZ}
        assert not cats.is_cat_to_sub_cat()

        # Act
        cats.set_cat_to_sub_cat({TestType.BAK: TestType.sub_2})

        # Assert
        assert cats.get_categories(as_dict=False, init=True) == [TestType.FOO, TestType.BAK, TestType.BAZ]
        assert cats.get_categories(as_dict=False) == [TestType.FOO, TestType.BAK_21, TestType.BAK_22, TestType.BAZ]
        assert cats.is_cat_to_sub_cat()

    @staticmethod
    @pytest.mark.basic
    def test_set_cat_to_subcat_and_check_categories_case_2() -> None:
        """
        Categories are dumped and annotation ids are correctly assigned and  is_cat_to_sub_cat works properly.
        """

        # Arrange
        cats = TestDatasetCategories.setup()

        # Act
        cats.set_cat_to_sub_cat({TestType.BAK: TestType.sub, TestType.FOO: TestType.cat})

        # Assert
        assert cats.get_categories(name_as_key=True) == {
            get_type("FOO_1"): "1",
            get_type("FOO_2"): "2",
            get_type("FOO_3"): "3",
            get_type("BAK_11"): "4",
            get_type("BAK_12"): "5",
            get_type("BAZ"): "6",
        }

    @staticmethod
    @pytest.mark.basic
    def test_filter_and_check_categories() -> None:
        """
        Categories are filtered and  is_filtered works properly
        """

        # Arrange
        cats = TestDatasetCategories.setup()

        # Assert
        assert not cats.is_filtered()

        # Act
        cats.set_cat_to_sub_cat({"BAK": "sub", "FOO": "cat"})
        cats.filter_categories(categories=["FOO_1", "BAZ", "BAK_11"])

        # Assert
        assert cats.get_categories(name_as_key=True, filtered=True) == {
            TestType.FOO_1: "1",
            TestType.BAZ: "3",
            TestType.BAK_11: "2",
        }
        assert cats.is_filtered()

    @staticmethod
    @pytest.mark.basic
    def test_check_sub_categories() -> None:
        """
        get_sub_categories works as expected
        """

        # Arrange
        cats = TestDatasetCategories.setup()

        # Assert
        assert cats.get_sub_categories() == {TestType.BAK: [TestType.sub, TestType.sub_2], TestType.FOO: [TestType.cat]}

    @staticmethod
    @pytest.mark.basic
    def test_set_sub_categories_and_check_sub_categories() -> None:
        """
        when categories are replaced with sub categories then get_sub_categories works as expected
        """

        # Arrange
        cats = TestDatasetCategories.setup()
        cats.set_cat_to_sub_cat({"FOO": "cat"})

        # Act
        assert cats.get_sub_categories(categories=["FOO_1", "FOO_2", "FOO_3"]) == {
            TestType.FOO_1: [],
            TestType.FOO_2: [],
            TestType.FOO_3: [],
        }


class TestMergeDatasetCategories:
    """
    Testing get_merged_categories
    """

    @staticmethod
    def setup() -> Tuple[DatasetCategories, DatasetCategories]:
        """
        Arrange testing setup
        """
        init_categories_1 = [get_type("FOO"), get_type("BAK")]
        init_categories_2 = [get_type("FOO"), get_type("BAZ")]

        sub_categories_1 = {
            get_type("FOO"): {
                get_type("FOO_1"): [get_type("1"), get_type("2")],
                get_type("FOO_2"): [get_type("3"), get_type("4")],
            }
        }
        sub_categories_2 = {
            get_type("FOO"): {get_type("FOO_1"): [get_type("1"), get_type("3")]},
            get_type("BAK"): {get_type("BAK_1"): [get_type("4"), get_type("5")]},
        }

        return (
            DatasetCategories(init_categories=init_categories_1, init_sub_categories=sub_categories_1),
            DatasetCategories(init_categories=init_categories_2, init_sub_categories=sub_categories_2),
        )

    @staticmethod
    @pytest.mark.basic
    def test_merge_categories_returns_union_categories_and_sub_categories() -> None:
        """
        Merge categories returns union of categories of datasets
        """

        # Arrange
        cat_1, cat_2 = TestMergeDatasetCategories.setup()

        # Act
        merge = get_merged_categories(cat_1, cat_2)

        # Assert
        assert merge.get_categories(init=True, as_dict=False) == [TestType.FOO, TestType.BAK, TestType.BAZ]
        assert merge.get_sub_categories() == {TestType.FOO: [TestType.FOO_1]}

    @staticmethod
    @pytest.mark.basic
    def test_merge_categories_updates_categories_correctly() -> None:
        """
        Merge categories updates categories when categories of datasets are being updated
        """

        # Arrange
        cat_1, cat_2 = TestMergeDatasetCategories.setup()
        cat_1.set_cat_to_sub_cat({"FOO": "FOO_1"})
        cat_2.set_cat_to_sub_cat({"FOO": "FOO_1"})

        # Act
        merge = get_merged_categories(cat_1, cat_2)

        # Assert
        assert merge.get_categories(as_dict=False, init=True) == [TestType.FOO, TestType.BAK, TestType.BAZ]
        assert merge.get_categories(as_dict=False) == [
            TestType.one,
            TestType.two,
            TestType.BAK,
            TestType.three,
            TestType.BAZ,
        ]

    @staticmethod
    @pytest.mark.basic
    def test_merge_categories_updates_and_filters_categories_correctly() -> None:
        """
        Merge categories returns updates and filtered categories correctly
        """

        # Arrange
        cat_1, cat_2 = TestMergeDatasetCategories.setup()
        cat_1.set_cat_to_sub_cat({"FOO": "FOO_1"})
        cat_1.filter_categories(categories=["2"])

        cat_2.set_cat_to_sub_cat({"FOO": "FOO_1"})

        # Act
        merge = get_merged_categories(cat_1, cat_2)

        # Assert
        assert merge.get_categories(as_dict=False, init=False) == [
            TestType.one,
            TestType.two,
            TestType.BAK,
            TestType.three,
            TestType.BAZ,
        ]
        assert merge.get_categories(as_dict=False, init=False, filtered=True) == [
            TestType.two,
            TestType.one,
            TestType.three,
            TestType.BAZ,
        ]

    @staticmethod
    @pytest.mark.basic
    def test_merge_categories_cannot_update_or_filter() -> None:
        """
        Calling :`filter_categories` or :`set_cat_to_sub_cat` is not allowed
        """

        # Arrange
        cat_1, cat_2 = TestMergeDatasetCategories.setup()
        merge = get_merged_categories(cat_1, cat_2)

        # Act & Assert
        with pytest.raises(PermissionError):
            merge.filter_categories(categories="BAZ")
            merge.set_cat_to_sub_cat({"FOO": "FOO_1"})



# -*- coding: utf-8 -*-
# File: tp_eval_callback.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for EvalCallback in Tensorpack
"""

from itertools import count
from typing import Mapping, Optional, Sequence, Type, Union

from ..datasets import DatasetBase
from ..extern.tpdetect import TPFrcnnDetector
from ..pipe.base import PredictorPipelineComponent
from ..utils.file_utils import tensorpack_available
from ..utils.logger import logger
from ..utils.metacfg import AttrDict
from ..utils.settings import ObjectTypes
from .base import MetricBase
from .eval import Evaluator

# pylint: disable=import-error
if tensorpack_available():
    from tensorpack.callbacks import Callback
    from tensorpack.predict import OnlinePredictor
    from tensorpack.utils.gpu import get_num_gpu
# pylint: enable=import-error


# The following class is modified from
# https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/eval.py

__all__ = ["EvalCallback"]


class EvalCallback(Callback):  # pylint: disable=R0903
    """
    A callback that runs evaluation once a while. It supports evaluation on any pipeline component.
    """

    _chief_only = False

    def __init__(
        self,
        dataset: DatasetBase,
        category_names: Optional[Union[ObjectTypes, Sequence[ObjectTypes]]],
        sub_categories: Optional[Union[Mapping[ObjectTypes, ObjectTypes], Mapping[ObjectTypes, Sequence[ObjectTypes]]]],
        metric: Union[Type[MetricBase], MetricBase],
        pipeline_component: PredictorPipelineComponent,
        in_names: str,
        out_names: str,
        **build_eval_kwargs: str,
    ) -> None:
        """
        :param dataset: dataset
        :param category_names: String or list of category names
        :param sub_categories: Dict of categories/sub-categories or categories/list of sub-categories. See also
                               `eval.Evaluator`
        :param metric: metric
        :param pipeline_component: Pipeline component with a detector.
        :param in_names: Specify tensor input names.
                         E.g. `extern.tp.tpfrcnn.GeneralizedRCNN.get_inference_tensor_names`
        :param out_names: Specify tensor output names.
        :param build_eval_kwargs: Pass the necessary arguments in order to build the dataflow, e.g. "split",
                                  "build_mode", "max_datapoints" etc.
        """
        self.dataset_name = dataset.dataset_info.name
        self.build_eval_kwargs = build_eval_kwargs
        self.in_names, self.out_names = in_names, out_names
        self.num_gpu = get_num_gpu()
        self.category_names = category_names
        self.sub_categories = sub_categories
        assert isinstance(pipeline_component.predictor, TPFrcnnDetector), (
            f"pipeline_component.predictor must be of "
            f"type TPFrcnnDetector but is type {type(pipeline_component.predictor)}"
        )
        self.cfg = pipeline_component.predictor.model.cfg
        if _use_replicated(self.cfg):
            self.evaluator = Evaluator(dataset, pipeline_component, metric, num_threads=self.num_gpu * 2)
        else:
            raise NotImplementedError("Can only evaluate in replicated training mode.")

    def _setup_graph(self) -> None:
        if _use_replicated(self.cfg):
            if self.evaluator.pipe_component is None:
                raise TypeError("self.evaluator.pipe_component cannot be None")
            for idx, comp in enumerate(self.evaluator.pipe_component.pipe_components):
                if not isinstance(comp, PredictorPipelineComponent):
                    raise TypeError(f"comp must be of type PredictorPipelineComponent but is type {type(comp)}")
                if not isinstance(comp.predictor, TPFrcnnDetector):
                    raise TypeError(
                        f"comp.predictor mus be of type TPFrcnnDetector but is of type {type(comp.predictor)}"
                    )
                comp.predictor.tp_predictor = self._build_predictor(idx % self.num_gpu)

    def _build_predictor(self, idx: int) -> OnlinePredictor:
        return self.trainer.get_predictor(self.in_names, self.out_names, device=idx)

    def _before_train(self) -> None:
        eval_period = self.cfg.TRAIN.EVAL_PERIOD
        self.epochs_to_eval = set()
        for k in count(1):
            if k * eval_period > self.trainer.max_epoch:
                break
            self.epochs_to_eval.add(k * eval_period)
        self.epochs_to_eval.add(self.trainer.max_epoch)
        logger.info("[EvalCallback] Will evaluate every %i epochs", eval_period)

    def _eval(self) -> None:
        scores = self.evaluator.run(True, **self.build_eval_kwargs)
        for k, val in scores.items():
            self.trainer.monitors.put_scalar(self.dataset_name + "-" + k, val)

    def _trigger_epoch(self) -> None:
        if self.epoch_num in self.epochs_to_eval:
            logger.info("Running evaluation ...")
            self._eval()


def _use_replicated(config: AttrDict) -> bool:
    if not hasattr(config, "TRAINER"):
        return False
    if config.TRAINER == "replicated":
        return True
    return False



# -*- coding: utf-8 -*-
# File: view.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Subclasses for ImageAnnotation and Image objects with various properties. These classes
simplify consumption
"""

import json
from copy import copy
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Type, Union, no_type_check

import cv2
import numpy as np

from ..utils.detection_types import ImageType, JsonDict, Pathlike
from ..utils.logger import logger
from ..utils.settings import CellType, LayoutType, ObjectTypes, PageType, Relationships, TableType, WordType, get_type
from ..utils.viz import draw_boxes, interactive_imshow
from .annotation import ContainerAnnotation, ImageAnnotation, SummaryAnnotation, ann_from_dict
from .box import BoundingBox
from .convert import convert_np_array_to_b64
from .image import Image


class ImageAnnotationBaseView(ImageAnnotation):
    """
    Consumption class for having easier access to categories added to an ImageAnnotation.

    ImageAnnotation is a generic class in the sense that different categories might have different
    sub categories collected while running through a pipeline. In order to get properties for a specific
    category one has to understand the internal data structure.

    To circumvent this obstacle `ImageAnnotationBaseView` provides the `__getattr__` so that
    to gather values defined by `ObjectTypes`. To be more precise: A sub class will have attributes either
    defined explicitly by a `@property` or by the set of `get_attribute_names()` . Do not define any attribute
    setter method and regard this class as a view to the super class.

    The class does contain its base page, which mean, that it is possible to retrieve all annotations that have a
    relation.

    base_page: `Page` class instantiated by the lowest hierarchy `Image`
    """

    base_page: "Page"

    @property
    def bbox(self) -> List[float]:
        """
        Get the bounding box as list and in absolute coordinates of the base page.
        """
        if self.image:
            bounding_box = self.image.get_embedding(self.base_page.image_id)
        else:
            bounding_box = self.bounding_box
        if not bounding_box.absolute_coords:
            bounding_box = bounding_box.transform(self.base_page.width, self.base_page.height, absolute_coords=True)
        return bounding_box.to_list(mode="xyxy")

    def __getattr__(self, item: str) -> Optional[Union[str, int, List[str]]]:
        """
        Get attributes defined by registered `self.get_attribute_names()` in a multi step process:

        - Unregistered attributes will raise an `AttributeError`.
        - Registered attribute will look for a corresponding sub category. If the sub category does not exist `Null`
          will be returned.
        - If the sub category exists it will return `category_name` provided that the attribute is not equal to the
          `category_name` otherwise
        - Check if the sub category is a `ContainerAnnotation` in which case the `value` will be returned otherwise
          `category_id` will be returned.
        - If nothing works, look at `self.image.summary` if the item exist. Follow the same logic as for ordinary sub
          categories.
        :param item: attribute name
        :return: value according to the logic described above
        """
        if item not in self.get_attribute_names():
            raise AttributeError(f"Attribute {item} is not supported for {type(self)}")
        if item in self.sub_categories:
            sub_cat = self.get_sub_category(get_type(item))
            if item != sub_cat.category_name:
                return sub_cat.category_name
            if isinstance(sub_cat, ContainerAnnotation):
                return sub_cat.value
            return int(sub_cat.category_id)
        if self.image is not None:
            if self.image.summary is not None:
                if item in self.image.summary.sub_categories:
                    sub_cat = self.image.summary.get_sub_category(get_type(item))
                    if item != sub_cat.category_name:
                        return sub_cat.category_name
                    if isinstance(sub_cat, ContainerAnnotation):
                        return sub_cat.value
                    return int(sub_cat.category_id)
        return None

    def get_attribute_names(self) -> Set[str]:  # pylint: disable=R0201
        """
        :return: A set of registered attributes. When sub classing modify this method accordingly.
        """
        return {"bbox"}

    @classmethod
    def from_dict(cls, **kwargs: JsonDict) -> "ImageAnnotationBaseView":
        """
        Identical to its base class method for having correct return types. If the base class changes, please
        change this method as well.
        """
        image_ann = ann_from_dict(cls, **kwargs)
        if box_kwargs := kwargs.get("bounding_box"):
            image_ann.bounding_box = BoundingBox.from_dict(**box_kwargs)
        return image_ann


class Word(ImageAnnotationBaseView):
    """
    Word specific subclass of `ImageAnnotationBaseView` modelled by `WordType`.
    """

    def get_attribute_names(self) -> Set[str]:
        return set(WordType).union(super().get_attribute_names()).union({Relationships.reading_order})


class Layout(ImageAnnotationBaseView):
    """
    Layout specific subclass of `ImageAnnotationBaseView`. In order check what ImageAnnotation will be wrapped
    into `Layout`, please consult `IMAGE_ANNOTATION_TO_LAYOUTS`.

    text_container: Pass the `LayoutObject` that is supposed to be used for `words`. It is possible that the
                    text_container is equal to `self.category_name`, in which case `words` returns `self`.
    """

    text_container: Optional[ObjectTypes] = None

    @property
    def words(self) -> List[ImageAnnotationBaseView]:
        """
        Get a list of `ImageAnnotationBaseView` objects with `LayoutType` defined by `text_container`.
        It will only select those among all annotations that have an entry in `Relationships.child` .
        """
        if self.category_name != self.text_container:
            text_ids = self.get_relationship(Relationships.child)
            return self.base_page.get_annotation(annotation_ids=text_ids, category_names=self.text_container)
        return [self]

    @property
    def text(self) -> str:
        """
        Text captured within the instance respecting the reading order of each word.
        """
        words_with_reading_order = [word for word in self.words if word.reading_order is not None]
        words_with_reading_order.sort(key=lambda x: x.reading_order)  # type: ignore
        return " ".join([word.characters for word in words_with_reading_order])  # type: ignore

    def get_attribute_names(self) -> Set[str]:
        return {"words", "text"}.union(super().get_attribute_names()).union({Relationships.reading_order})

    def __len__(self) -> int:
        """len of text counted by number of characters"""
        return len(self.text)


class Cell(Layout):
    """
    Cell specific subclass of `ImageAnnotationBaseView` modelled by `CellType`.
    """

    def get_attribute_names(self) -> Set[str]:
        return set(CellType).union(super().get_attribute_names())


class Table(Layout):
    """
    Table specific sub class of `ImageAnnotationBaseView` modelled by `TableType`.
    """

    @property
    def cells(self) -> List[ImageAnnotationBaseView]:
        """
        A list of a table cells.
        """
        all_relation_ids = self.get_relationship(Relationships.child)
        cell_anns = self.base_page.get_annotation(
            annotation_ids=all_relation_ids,
            category_names=[
                LayoutType.cell,
                CellType.header,
                CellType.body,
                CellType.projected_row_header,
                CellType.spanning,
                CellType.row_header,
                CellType.column_header,
            ],
        )
        return cell_anns

    @property
    def rows(self) -> List[ImageAnnotationBaseView]:
        """
        A list of a table rows.
        """
        all_relation_ids = self.get_relationship(Relationships.child)
        row_anns = self.base_page.get_annotation(annotation_ids=all_relation_ids, category_names=[LayoutType.row])
        return row_anns

    @property
    def columns(self) -> List[ImageAnnotationBaseView]:
        """
        A list of a table columns.
        """
        all_relation_ids = self.get_relationship(Relationships.child)
        col_anns = self.base_page.get_annotation(annotation_ids=all_relation_ids, category_names=[LayoutType.column])
        return col_anns

    @property
    def html(self) -> str:
        """
        The html representation of the table
        """

        html_list = []
        if TableType.html in self.sub_categories:
            ann = self.get_sub_category(TableType.html)
            if isinstance(ann, ContainerAnnotation):
                if isinstance(ann.value, list):
                    html_list = copy(ann.value)
        for cell in self.cells:
            try:
                html_index = html_list.index(cell.annotation_id)
                html_list.pop(html_index)
                html_list.insert(html_index, cell.text)  # type: ignore
            except ValueError:
                logger.warning("html construction not possible due to ValueError in: %s", cell.annotation_id)

        return "".join(html_list)

    def get_attribute_names(self) -> Set[str]:
        return set(TableType).union(super().get_attribute_names()).union({"cells", "rows", "columns", "html"})

    @property
    def csv(self) -> List[List[str]]:
        """Returns a csv-style representation of a table as list of lists of string. Cell content of cell with higher
        row or column spans will be shown at the upper left cell tile. All other tiles covered by the cell will be left
        as blank
        """
        cells = self.cells
        table_list = [["" for _ in range(self.number_of_columns)] for _ in range(self.number_of_rows)]  # type: ignore
        for cell in cells:
            table_list[cell.row_number - 1][cell.column_number - 1] = cell.text  # type: ignore
        return table_list

    def __str__(self) -> str:
        out = " ".join([" ".join(row + ["\n"]) for row in self.csv])
        return out


IMAGE_ANNOTATION_TO_LAYOUTS: Dict[ObjectTypes, Type[Union[Layout, Table, Word]]] = {
    **{i: Layout for i in LayoutType if (i not in {LayoutType.table, LayoutType.word, LayoutType.cell})},
    LayoutType.table: Table,
    LayoutType.word: Word,
    LayoutType.cell: Cell,
    CellType.projected_row_header: Cell,
    CellType.spanning: Cell,
    CellType.row_header: Cell,
    CellType.column_header: Cell,
}


@no_type_check
def ann_obj_view_factory(annotation: ImageAnnotation, text_container: ObjectTypes) -> ImageAnnotationBaseView:
    """
    Create an `ImageAnnotationBaseView` sub class given the mapping `IMAGE_ANNOTATION_TO_LAYOUTS` .

    :param annotation: The annotation to transform. Note, that we do not use the input annotation as base class
                       but create a whole new instance.
    :param text_container: `LayoutType` to create a list of `words` and eventually generate `text`
    :return: Transformed annotation
    """

    # We need to handle annotations that are text containers like words
    if annotation.category_name == text_container:
        layout_class = IMAGE_ANNOTATION_TO_LAYOUTS[LayoutType.word]
    else:
        layout_class = IMAGE_ANNOTATION_TO_LAYOUTS[annotation.category_name]
    ann_dict = annotation.as_dict()
    layout = layout_class.from_dict(**ann_dict)
    if image_dict := ann_dict.get("image"):
        layout.image = Page.from_dict(**image_dict)
    layout.text_container = text_container
    return layout


class Page(Image):
    """
    Consumer class for its super `Image` class. It comes with some handy `@property` as well as
    custom `__getattr__` to give easier access to various information that are stored in the base class
    as `ImageAnnotation` or `CategoryAnnotation`.

    Its factory function `Page().from_image(image, text_container, text_block_names)` creates for every
    `ImageAnnotation` a corresponding subclass of `ImageAnnotationBaseView` which drives the object towards
    less generic classes with custom attributes that are controlled some `ObjectTypes`.

    top_level_text_block_names: Top level layout objects, e.g. `LayoutType.text` or `LayoutType.table`.

    text_block_names: layout objects that have associated text

    image_orig: Base image

    text_container: LayoutType to take the text from
    """

    top_level_text_block_names: List[ObjectTypes]
    text_block_names: Optional[List[ObjectTypes]]
    image_orig: Image
    text_container: ObjectTypes

    @no_type_check
    def get_annotation(
        self,
        category_names: Optional[Union[str, ObjectTypes, Sequence[Union[str, ObjectTypes]]]] = None,
        annotation_ids: Optional[Union[str, Sequence[str]]] = None,
        annotation_types: Optional[Union[str, Sequence[str]]] = None,
    ) -> List[ImageAnnotationBaseView]:
        """
        Identical to its base class method for having correct return types. If the base class changes, please
        change this method as well.
        """
        cat_names = [category_names] if isinstance(category_names, (ObjectTypes, str)) else category_names
        if cat_names is not None:
            cat_names = [get_type(cat_name) for cat_name in cat_names]
        ann_ids = [annotation_ids] if isinstance(annotation_ids, str) else annotation_ids
        ann_types = [annotation_types] if isinstance(annotation_types, str) else annotation_types

        anns = filter(lambda x: x.active, self.annotations)

        if ann_types is not None:
            for type_name in ann_types:
                anns = filter(lambda x: isinstance(x, eval(type_name)), anns)  # pylint: disable=W0123, W0640

        if cat_names is not None:
            anns = filter(lambda x: x.category_name in cat_names, anns)

        if ann_ids is not None:
            anns = filter(lambda x: x.annotation_id in ann_ids, anns)

        return list(anns)

    def __getattr__(self, item: str) -> Any:
        if item not in self.get_attribute_names():
            raise AttributeError(f"Attribute {item} is not supported for {type(self)}")
        if self.summary is not None:
            if item in self.summary.sub_categories:
                sub_cat = self.summary.get_sub_category(get_type(item))
                if item != sub_cat.category_name:
                    return sub_cat.category_name
                if isinstance(sub_cat, ContainerAnnotation):
                    return sub_cat.value
                return int(sub_cat.category_id)
        return None

    @property
    def layouts(self) -> List[ImageAnnotationBaseView]:
        """
        A list of a layouts.
        """
        layouts = [layout for layout in self.top_level_text_block_names if layout != LayoutType.table]
        return self.get_annotation(category_names=layouts)

    @property
    def words(self) -> List[ImageAnnotationBaseView]:
        """
        A list of a words.
        """
        return self.get_annotation(category_names=self.text_container)

    @property
    def residual_words(self) -> List[ImageAnnotationBaseView]:
        """
        A list of a words that have not been assigned to any text block but have a reading order.
        Words having this property appear, once `text_containers_to_text_block=True`.
        """
        if self.text_block_names is None:
            return []
        text_block_anns = self.get_annotation(category_names=self.text_block_names)
        text_ann_ids = list(
            chain(*[text_block.get_relationship(Relationships.child) for text_block in text_block_anns])
        )
        text_container_anns = self.get_annotation(category_names=self.text_container)
        residual_words = [ann for ann in text_container_anns if ann.annotation_id not in text_ann_ids]
        return residual_words

    @property
    def tables(self) -> List[ImageAnnotationBaseView]:
        """
        A list of a tables.
        """
        return self.get_annotation(category_names=LayoutType.table)

    @classmethod
    def from_image(
        cls,
        image_orig: Image,
        text_container: ObjectTypes,
        top_level_text_block_names: List[ObjectTypes],
        text_block_names: Optional[List[ObjectTypes]] = None,
        base_page: Optional["Page"] = None,
    ) -> "Page":
        """
        Factory function for generating a `Page` instance from `image_orig` .

        :param image_orig: `Image` instance to convert
        :param text_container: A LayoutType to get the text from. It will steer the output of `Layout.words`.
        :param top_level_text_block_names: A list of top level layout objects
        :param text_block_names: name of image annotation that have a relation with text containers (or which might be
                                 text containers themselves). This is only necessary, when residual text_container (e.g.
                                 words that have not been assigned to any text block) should be displayed in `page.text`
        :param base_page: For top level objects that are images themselves, pass the page that encloses all objects.
                          In doubt, do not populate this value.
        :return:
        """
        img_kwargs = image_orig.as_dict()
        page = cls(
            img_kwargs.get("file_name"), img_kwargs.get("location"), img_kwargs.get("external_id")  # type: ignore
        )
        page.image_orig = image_orig
        page.page_number = image_orig.page_number
        page.document_id = image_orig.document_id
        if image_orig.image is not None:
            page.image = image_orig.image  # pass image explicitly so
        page._image_id = img_kwargs.get("_image_id")
        if page.image is None:
            if b64_image := img_kwargs.get("_image"):
                page.image = b64_image
        if box_kwargs := img_kwargs.get("_bbox"):
            page._bbox = BoundingBox.from_dict(**box_kwargs)
        if embeddings := img_kwargs.get("embeddings"):
            for image_id, box_dict in embeddings.items():
                page.set_embedding(image_id, BoundingBox.from_dict(**box_dict))
        for ann_dict in img_kwargs.get("annotations", []):
            image_ann = ImageAnnotation.from_dict(**ann_dict)
            layout_ann = ann_obj_view_factory(image_ann, text_container)
            if "image" in ann_dict:
                image_dict = ann_dict["image"]
                if image_dict:
                    image = Image.from_dict(**image_dict)
                    layout_ann.image = cls.from_image(
                        image, text_container, top_level_text_block_names, text_block_names, page
                    )
            layout_ann.base_page = base_page if base_page is not None else page
            page.dump(layout_ann)
        if summary_dict := img_kwargs.get("_summary"):
            page.summary = SummaryAnnotation.from_dict(**summary_dict)
        page.top_level_text_block_names = top_level_text_block_names
        page.text_block_names = text_block_names
        page.text_container = text_container
        return page

    def _order(self, block: str) -> List[ImageAnnotationBaseView]:
        blocks_with_order = [layout for layout in getattr(self, block) if layout.reading_order is not None]
        if self.residual_words:
            blocks_with_order.extend(self.residual_words)
        blocks_with_order.sort(key=lambda x: x.reading_order)
        return blocks_with_order

    @property
    def text(self) -> str:
        """
        Get text of all layouts.
        """
        text: str = ""
        block_name = "layouts" if self.layouts else "words"
        block_with_order = self._order(block_name)
        linebreak = "\n" if block_name == "layouts" else " "
        for block in block_with_order:
            block_attr = "text" if block.category_name != LayoutType.word else "characters"
            text += f"{linebreak}{getattr(block, block_attr)}"
        return text

    @property
    def chunks(self) -> List[Tuple[str, str, str, str, str, str]]:
        """
        :return: Returns a "chunk" of a layout element or a table as 6-tuple containing

                    - document id
                    - image id
                    - reading order
                    - category name
                    - text string

        """
        block_with_order = self._order("layouts")
        for table in self.tables:
            if table.reading_order:
                block_with_order.append(table)
        all_chunks = []
        for chunk in block_with_order:
            all_chunks.append(
                (
                    self.document_id,
                    self.image_id,
                    chunk.annotation_id,
                    chunk.reading_order,
                    chunk.category_name,
                    chunk.text,
                )
            )
        return all_chunks  # type: ignore

    @property
    def text_no_line_break(self) -> str:
        """
        Get text of all layouts. While `text` will do a line break for each layout block this here will return the
        string in one single line.
        """
        text: str = ""
        layouts_with_order = self._order("layouts")
        for layout in layouts_with_order:
            text += " " + layout.text  # type: ignore

        return text

    @no_type_check
    def viz(
        self,
        show_tables: bool = True,
        show_layouts: bool = True,
        show_cells: bool = True,
        show_table_structure: bool = True,
        show_words: bool = False,
        show_token_class: bool = True,
        interactive: bool = False,
    ) -> Optional[ImageType]:
        """
        Display a page detected bounding boxes. One can select bounding boxes of tables or other layout components.

        **Example:**

                from matplotlib import pyplot as plt

                img = page.viz()
                plt.imshow(img)

        :param show_tables: Will display all tables boxes as well as cells, rows and columns
        :param show_layouts: Will display all other layout components.
        :param show_cells: Will display cells within tables. (Only available if `show_tables=True`)
        :param show_table_structure: Will display rows and columns
        :param show_words: Will display bounding boxes around words labeled with token class and bio tag (experimental)
        :param show_token_class: Will display token class instead of token tags (i.e. token classes with tags)
        :param interactive: If set to True will open an interactive image, otherwise it will return a numpy array that
                            can be displayed differently.
        :return: If interactive will return nothing else a numpy array.
        """

        category_names_list: List[Union[str, None]] = []
        box_stack = []
        cells_found = False

        if show_layouts:
            for item in self.layouts:
                box_stack.append(item.bbox)
                category_names_list.append(item.category_name)

        if show_tables:
            for table in self.tables:
                box_stack.append(table.bbox)
                category_names_list.append(LayoutType.table)
                if show_cells:
                    for cell in table.cells:
                        if cell.category_name != LayoutType.cell:
                            cells_found = True
                            box_stack.append(cell.bbox)
                            category_names_list.append(None)
                if show_table_structure:
                    rows = table.rows
                    cols = table.columns
                    for row in rows:
                        box_stack.append(row.bbox)
                        category_names_list.append(None)
                    for col in cols:
                        box_stack.append(col.bbox)
                        category_names_list.append(None)

        if show_cells and not cells_found:
            for ann in self.annotations:
                if isinstance(ann, Cell):
                    box_stack.append(ann.bbox)
                    category_names_list.append(None)

        if show_words:
            all_words = []
            for layout in self.layouts:
                all_words.extend(layout.words)
            for word in all_words:
                box_stack.append(word.bbox)
                if show_token_class:
                    category_names_list.append(str(word.token_class).replace("TokenClasses", ""))
                else:
                    category_names_list.append(str(word.token_tag))

        if self.image is not None:
            if box_stack:
                boxes = np.vstack(box_stack)
                if show_words:
                    img = draw_boxes(
                        self.image,
                        boxes,
                        category_names_list,
                        color=(255, 222, 173),
                        font_scale=0.25,
                        rectangle_thickness=1,
                    )
                else:
                    img = draw_boxes(self.image, boxes, category_names_list)
                img = cv2.resize(img, None, fx=1.3, fy=1.3, interpolation=cv2.INTER_CUBIC)
            else:
                img = self.image

            if interactive:
                interactive_imshow(img)
                return None
            return img
        return None

    @staticmethod
    def get_attribute_names() -> Set[str]:
        """
        :return: A set of registered attributes.
        """
        return set(PageType).union({"text", "tables", "layouts", "words", "residual_words"})

    def save(
        self,
        image_to_json: bool = True,
        highest_hierarchy_only: bool = False,
        path: Optional[Pathlike] = None,
        dry: bool = False,
    ) -> Optional[JsonDict]:
        """
        Export image as dictionary. As numpy array cannot be serialized `image` values will be converted into
        base64 encodings.
        :param image_to_json: If True will save the image as b64 encoded string in output
        :param highest_hierarchy_only: If True it will remove all image attributes of ImageAnnotations
        :param path: Path to save the .json file to
        :param dry: Will run dry, i.e. without saving anything but returning the dict

        :return: optional dict
        """
        if isinstance(path, str):
            path = Path(path)
        elif path is None:
            path = Path(self.image_orig.location)
        suffix = path.suffix
        path_json = path.as_posix().replace(suffix, ".json")
        if highest_hierarchy_only:
            self.image_orig.remove_image_from_lower_hierachy()
        export_dict = self.image_orig.as_dict()
        export_dict["location"] = str(export_dict["location"])
        if image_to_json and self.image_orig.image is not None:
            export_dict["_image"] = convert_np_array_to_b64(self.image_orig.image)
        if dry:
            return export_dict
        with open(path_json, "w", encoding="UTF-8") as file:
            json.dump(export_dict, file)
        return None



# -*- coding: utf-8 -*-
# File: generalized_rcnn.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file is modified from
<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/modeling/generalized_rcnn.py>
"""

# pylint: disable=import-error
import tensorflow as tf
from tensorpack import tfv1
from tensorpack.models import l2_regularizer, regularize_cost
from tensorpack.tfutils import optimizer
from tensorpack.tfutils.summary import add_moving_summary

from ...tpcompat import ModelDescWithConfig
from ..utils.box_ops import area as tf_area
from . import model_frcnn, model_mrcnn
from .backbone import image_preprocess, resnet_fpn_backbone
from .model_box import RPNAnchors, clip_boxes, crop_and_resize
from .model_cascade import CascadeRCNNHead
from .model_fpn import (
    fpn_model,
    generate_fpn_proposals,
    get_all_anchors_fpn,
    multilevel_roi_align,
    multilevel_rpn_losses,
)
from .model_frcnn import (
    BoxProposals,
    FastRCNNHead,
    fastrcnn_outputs,
    fastrcnn_predictions,
    nms_post_processing,
    sample_fast_rcnn_targets,
)
from .model_mrcnn import maskrcnn_loss, unpackbits_masks
from .model_rpn import rpn_head


class GeneralizedRCNN(ModelDescWithConfig):
    """
    GeneralizedRCNN
    """

    def preprocess(self, image):
        """
        Pre-processing steps

        :param image: tf.Tensor
        :return: tf.Tensor
        """
        image = tf.expand_dims(image, 0)
        image = image_preprocess(image, self.cfg)
        return tf.transpose(image, [0, 3, 1, 2])

    def optimizer(self):
        """
        optimizer
        """
        learning_rate = tfv1.get_variable("learning_rate", initializer=0.003, trainable=False)
        tf.summary.scalar("learning_rate-summary", learning_rate)

        # The learning rate in the config is set for 8 GPUs, and we use trainers with average=False.
        learning_rate = learning_rate / 8.0
        opt = tfv1.train.MomentumOptimizer(learning_rate, 0.9)
        if self.cfg.TRAIN.NUM_GPUS < 8:
            opt = optimizer.AccumGradOptimizer(opt, 8 // self.cfg.TRAIN.NUM_GPUS)
        return opt

    def get_inference_tensor_names(self):
        """
        Returns two lists of tensor names to be used to create an inference callable.

        `build_graph` must create tensors of these names when called under inference context.
        :return: input names as list of strings, output names as list of strings
        """

        out = ["output/boxes", "output/scores", "output/labels"]
        if self.cfg.MODE_MASK:
            out.append("output/masks")
        return ["image"], out

    def build_graph(self, *inputs):  # pylint: disable=R1710
        """
        Build the graph

        :param inputs: args of inputs as defined in the self.inputs of subclass
        :return: total loss or None depending on training mode
        """
        inputs = dict(zip(self.input_names, inputs))
        if "gt_masks_packed" in inputs:
            gt_masks = tf.cast(unpackbits_masks(inputs.pop("gt_masks_packed")), tf.uint8, name="gt_masks")
            inputs["gt_masks"] = gt_masks

        image = self.preprocess(inputs["image"])  # 1CHW

        features = self.backbone(image)
        anchor_inputs = {k: v for k, v in inputs.items() if k.startswith("anchor_")}
        proposals, rpn_losses = self.rpn(image, features, anchor_inputs)

        targets = [inputs[k] for k in ["gt_boxes", "gt_labels", "gt_masks"] if k in inputs]
        gt_boxes_area = tf.reduce_mean(tf_area(inputs["gt_boxes"]), name="mean_gt_box_area")
        add_moving_summary(gt_boxes_area)
        head_losses = self.roi_heads(image, features, proposals, targets)

        if self.training:
            wd_cost = regularize_cost(".*/W", l2_regularizer(self.cfg.TRAIN.WEIGHT_DECAY), name="wd_cost")
            total_cost = tf.add_n(rpn_losses + head_losses + [wd_cost], "total_cost")
            add_moving_summary(total_cost, wd_cost)
            return total_cost

        # Check that the model defines the tensors it declares for inference
        # For existing models, they are defined in "fastrcnn_predictions(name_scope='output')"
        G = tf.compat.v1.get_default_graph()
        n_s = G.get_name_scope()
        for name in self.get_inference_tensor_names()[1]:
            try:
                name = "/".join([n_s, name]) if n_s else name
                G.get_tensor_by_name(name + ":0")
            except KeyError:
                raise KeyError(  # pylint: disable=W0707
                    f"Your model does not define the tensor '{name}' in inference context."
                )


class ResNetFPNModel(GeneralizedRCNN):
    """
    FPN and Cascade-RCNN with resnet/resnext backbone options
    """

    def inputs(self):
        """
        inputs

        :return: List of TensorSpec
        """
        ret = [tf.TensorSpec((None, None, 3), tf.float32, "image")]
        num_anchors = len(self.cfg.RPN.ANCHOR_RATIOS)
        for k in range(len(self.cfg.FPN.ANCHOR_STRIDES)):
            ret.extend(
                [
                    tf.TensorSpec((None, None, num_anchors), tf.int32, f"anchor_labels_lvl{k + 2}"),
                    tf.TensorSpec((None, None, num_anchors, 4), tf.float32, f"anchor_boxes_lvl{k+2}"),
                ]
            )
        ret.extend(
            [tf.TensorSpec((None, 4), tf.float32, "gt_boxes"), tf.TensorSpec((None,), tf.int64, "gt_labels")]
        )  # all > 0
        if self.cfg.MODE_MASK:
            ret.append(tf.TensorSpec((None, None, None), tf.uint8, "gt_masks_packed"))
        return ret

    def slice_feature_and_anchors(self, p23456, anchors):
        """
        slice_feature_and_anchors

        :param p23456: output from fpn backbone
        :param anchors: anchors
        """
        for i, _ in enumerate(self.cfg.FPN.ANCHOR_STRIDES):
            with tf.name_scope(f"FPN_slice_lvl{i}"):
                anchors[i] = anchors[i].narrow_to(p23456[i])

    def backbone(self, image):
        """
        backbone

        :param image: tf.Tensor
        """
        c2345 = resnet_fpn_backbone(image, self.cfg)
        p23456 = fpn_model("fpn", c2345, self.cfg.FPN.NUM_CHANNEL, self.cfg.FPN.NORM)  # pylint: disable=E1121
        return p23456

    def rpn(self, image, features, inputs):
        """
        Region Proposal Network

        :param image: tf.Tensor
        :param features: output features
        :param inputs: input dict of anchor levels
        :return: BoxProposals and list of losses
        """
        assert len(self.cfg.RPN.ANCHOR_SIZES) == len(self.cfg.FPN.ANCHOR_STRIDES)

        image_shape2d = tf.shape(image)[2:]  # h,w
        all_anchors_fpn = get_all_anchors_fpn(
            strides=self.cfg.FPN.ANCHOR_STRIDES,
            sizes=self.cfg.RPN.ANCHOR_SIZES,
            ratios=self.cfg.RPN.ANCHOR_RATIOS,
            max_size=self.cfg.PREPROC.MAX_SIZE,
        )
        multilevel_anchors = [
            RPNAnchors(
                all_anchors_fpn[i],
                inputs[f"anchor_labels_lvl{i+2}"],
                inputs[f"anchor_boxes_lvl{i+2}"],
            )
            for i in range(len(all_anchors_fpn))
        ]
        self.slice_feature_and_anchors(features, multilevel_anchors)

        # Multi-Level RPN Proposals
        rpn_outputs = [
            rpn_head("rpn", pi, self.cfg.FPN.NUM_CHANNEL, len(self.cfg.RPN.ANCHOR_RATIOS))  # pylint: disable=E1121
            for pi in features
        ]
        multilevel_label_logits = [k[0] for k in rpn_outputs]
        multilevel_box_logits = [k[1] for k in rpn_outputs]
        multilevel_pred_boxes = [
            anchor.decode_logits(logits, self.cfg.PREPROC.MAX_SIZE)
            for anchor, logits in zip(multilevel_anchors, multilevel_box_logits)
        ]

        proposal_boxes, _ = generate_fpn_proposals(
            multilevel_pred_boxes,
            multilevel_label_logits,
            image_shape2d,
            self.cfg.FPN.ANCHOR_STRIDES,
            self.cfg.FPN.PROPOSAL_MODE,
            self.cfg.RPN.TRAIN_PER_LEVEL_NMS_TOPK,
            self.cfg.RPN.PER_LEVEL_NMS_TOPK,
            self.cfg.RPN.MIN_SIZE,
            self.cfg.RPN.PROPOSAL_NMS_THRESH,
            self.cfg.RPN.TRAIN_PRE_NMS_TOPK,
            self.cfg.RPN.PRE_NMS_TOPK,
            self.cfg.RPN.TRAIN_POST_NMS_TOPK,
            self.cfg.RPN.POST_NMS_TOPK,
        )

        if self.training:
            losses = multilevel_rpn_losses(
                multilevel_anchors,
                multilevel_label_logits,
                multilevel_box_logits,
                self.cfg.RPN.BATCH_PER_IM,
                self.cfg.FPN.ANCHOR_STRIDES,
            )
        else:
            losses = []

        return BoxProposals(proposal_boxes), losses

    def roi_heads(self, image, features, proposals, targets):
        """
        Region of Interest head classifier and regressor

        :param image: tf.Tensor
        :param features: features from backbone output
        :param proposals: RPN proposals
        :param targets: targets
        :return: list of total loss when training or an empty list
        """
        image_shape2d = tf.shape(image)[2:]  # h,w
        assert len(features) == 5, "Features have to be P23456!"
        gt_boxes, gt_labels, *_ = targets

        if self.training:
            proposals = sample_fast_rcnn_targets(
                proposals.boxes,
                gt_boxes,
                gt_labels,
                self.cfg.FRCNN.FG_THRESH,
                self.cfg.FRCNN.BATCH_PER_IM,
                self.cfg.FRCNN.FG_RATIO,
            )

        fastrcnn_head_func = getattr(model_frcnn, self.cfg.FPN.FRCNN_HEAD_FUNC)
        if not self.cfg.FPN.CASCADE:
            roi_feature_fastrcnn = multilevel_roi_align(features[:4], proposals.boxes, 7, self.cfg.FPN.ANCHOR_STRIDES)

            head_feature = fastrcnn_head_func("fastrcnn", roi_feature_fastrcnn, cfg=self.cfg)
            fastrcnn_label_logits, fastrcnn_box_logits = fastrcnn_outputs(
                "fastrcnn/outputs", head_feature, self.cfg.DATA.NUM_CATEGORY
            )
            fastrcnn_head = FastRCNNHead(
                proposals,
                fastrcnn_box_logits,
                fastrcnn_label_logits,
                gt_boxes,
                tf.constant(self.cfg.FRCNN.BBOX_REG_WEIGHTS, dtype=tf.float32),
                self.cfg,
            )
        else:

            def roi_func(boxes):
                return multilevel_roi_align(features[:4], boxes, 7, self.cfg.FPN.ANCHOR_STRIDES)

            fastrcnn_head = CascadeRCNNHead(
                proposals,
                roi_func,
                fastrcnn_head_func,
                (gt_boxes, gt_labels),
                image_shape2d,
                self.cfg.DATA.NUM_CATEGORY,
                self.cfg,
            )

        if self.training:
            all_losses = fastrcnn_head.losses()

            if self.cfg.MODE_MASK:
                gt_masks = targets[2]
                # maskrcnn loss
                roi_feature_maskrcnn = multilevel_roi_align(
                    features[:4],
                    proposals.fg_boxes(),
                    14,
                    self.cfg,
                    name_scope="multilevel_roi_align_mask",
                )
                maskrcnn_head_func = getattr(model_mrcnn, self.cfg.FPN.MRCNN_HEAD_FUNC)
                mask_logits = maskrcnn_head_func(
                    "maskrcnn", roi_feature_maskrcnn, self.cfg.DATA.NUM_CATEGORY, cfg=self.cfg
                )  # #fg x #cat x 28 x 28

                target_masks_for_fg = crop_and_resize(
                    tf.expand_dims(gt_masks, 1), proposals.fg_boxes(), proposals.fg_inds_wrt_gt, 28, pad_border=False
                )  # fg x 1x28x28
                target_masks_for_fg = tf.squeeze(target_masks_for_fg, 1, "sampled_fg_mask_targets")
                all_losses.append(maskrcnn_loss(mask_logits, proposals.fg_labels(), target_masks_for_fg))
            return all_losses
        decoded_boxes = fastrcnn_head.decoded_output_boxes()
        decoded_boxes = clip_boxes(decoded_boxes, image_shape2d, name="fastrcnn_all_boxes")
        label_scores = fastrcnn_head.output_scores(name="fastrcnn_all_scores")
        boxes, scores, labels = fastrcnn_predictions(
            decoded_boxes,
            label_scores,
            self.cfg.OUTPUT.RESULT_SCORE_THRESH,
            self.cfg.OUTPUT.RESULTS_PER_IM,
            self.cfg.OUTPUT.FRCNN_NMS_THRESH,
            name_scope="pre_output",
        )
        final_boxes, _, final_labels = nms_post_processing(
            boxes,
            scores,
            labels,
            self.cfg.OUTPUT.RESULTS_PER_IM,
            self.cfg.OUTPUT.NMS_THRESH_CLASS_AGNOSTIC,
            name_scope="output",
        )

        if self.cfg.MODE_MASK:
            # Cascade inference needs roi transform with refined boxes.
            roi_feature_maskrcnn = multilevel_roi_align(features[:4], final_boxes, 14)  # pylint: disable=E1120
            maskrcnn_head_func = getattr(model_mrcnn, self.cfg.FPN.MRCNN_HEAD_FUNC)
            mask_logits = maskrcnn_head_func(
                "maskrcnn", roi_feature_maskrcnn, self.cfg.DATA.NUM_CATEGORY
            )  # #fg x #cat x 28 x 28
            indices = tf.stack([tf.range(tf.size(final_labels)), tf.cast(final_labels, tf.int32) - 1], axis=1)
            final_mask_logits = tf.gather_nd(mask_logits, indices)  # #resultx28x28
            tf.sigmoid(final_mask_logits, name="output/masks")
        return []



# -*- coding: utf-8 -*-
# File: test_misc.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module mapper.misc
"""
from typing import Dict, Union
from unittest.mock import MagicMock, patch

from numpy import ones, uint8
from numpy.testing import assert_array_equal
from pytest import mark

from deepdoctection.datapoint import Image
from deepdoctection.mapper.misc import to_image

_TEST_IMAGE = Image(file_name="test_image.png", location="test/to/path/test_image.png")
_TEST_IMAGE.image = ones((4, 3, 3), dtype=uint8)

_TEST_IMAGE_2 = Image(file_name="test_image.pdf", location="test/to/path/test_image.pdf")
_TEST_IMAGE_2.image = ones((4, 3, 3), dtype=uint8)


@mark.basic
@mark.parametrize(
    "datapoint,expected_image",
    [
        ("test/to/path/test_image.png", _TEST_IMAGE),
        ({"file_name": "test_image.png", "location": "test/to/path/test_image.png"}, _TEST_IMAGE),
        ({"file_name": "test_image.png", "path": "test/to/path/test_image.png"}, _TEST_IMAGE),
        (
            {"file_name": "test_image.pdf", "path": "test/to/path/test_image.pdf", "pdf_bytes": b"some_bytes"},
            _TEST_IMAGE_2,
        ),
    ],
)
@patch("deepdoctection.mapper.misc.load_image_from_file", MagicMock(return_value=ones((4, 3, 3))))
@patch("deepdoctection.mapper.misc.convert_pdf_bytes_to_np_array_v2", MagicMock(return_value=ones((4, 3, 3))))
def test_to_image(datapoint: Union[str, Dict[str, Union[str, bytes]]], expected_image: Image) -> None:
    """
    Image is properly constructed
    :param datapoint: input to testing method
    :param expected_image: expected Image output
    """

    # Act

    dp = to_image(datapoint)

    # Assert
    assert dp is not None
    assert dp.image_id == expected_image.image_id
    assert dp.file_name == expected_image.file_name
    assert dp.location == expected_image.location
    assert_array_equal(dp.get_image().to_np_array(), expected_image.image)  # type: ignore

    if dp.file_name.endswith("pdf"):
        assert dp.pdf_bytes == datapoint["pdf_bytes"]  # type: ignore



# -*- coding: utf-8 -*-
# File: __init__.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Init file for instances package. Place all datasets in a directory using the structure

    deepdoctection
    ├── datasets
    │ ├── dataset_1
    │ ├── dataset_2
    │ ├── dataset_3

If not sure use

    print(dataset_instance.dataflow.get_workdir())
"""

from .doclaynet import *
from .fintabnet import *
from .funsd import *
from .iiitar13k import *
from .layouttest import *
from .publaynet import *
from .pubtables1m import *
from .pubtabnet import *
from .rvlcdip import *
from .xfund import *

__all__ = [
    "Publaynet",
    "Pubtabnet",
    "LayoutTest",
    "Fintabnet",
    "Xfund",
    "Funsd",
    "IIITar13K",
    "Pubtables1MDet",
    "Pubtables1MStruct",
    "Rvlcdip",
    "DocLayNet",
]



# -*- coding: utf-8 -*-
# File: test_image.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module datapoint.image
"""

from typing import Union

from numpy import float32, ones
from numpy.testing import assert_array_equal
from pytest import mark, raises

from deepdoctection.dataflow import DataFlow, MapData, SerializerJsonlines
from deepdoctection.datapoint import BoundingBox, CategoryAnnotation, Image, ImageAnnotation
from deepdoctection.utils import get_uuid
from deepdoctection.utils.settings import get_type

from ..test_utils import anns_to_ids, collect_datapoint_from_dataflow, get_test_path
from .conftest import TestPdfPage, WhiteImage


class TestImage:
    """
    Testing Image methods
    """

    @staticmethod
    @mark.basic
    @mark.parametrize(
        "location, file_name, external_id, expected",
        [
            (WhiteImage.loc, WhiteImage.file_name, None, WhiteImage.get_image_id("d")),
            (
                WhiteImage.loc,
                WhiteImage.file_name,
                WhiteImage.external_id,
                WhiteImage.get_image_id("n"),
            ),
            (WhiteImage.loc, WhiteImage.file_name, WhiteImage.uuid, WhiteImage.get_image_id("u")),
        ],
    )
    def test_image_id_is_correctly_assigned(
        location: str, file_name: str, external_id: Union[str, None], expected: str
    ) -> None:
        """
        Image_id is assigned as expected, no matter what external_id is provided
        :param location: loc
        :param file_name: file name
        :param external_id: external id
        :param expected: image_id
        :return:
        """
        # Arrange
        test_image = Image(file_name=file_name, location=location, external_id=external_id)

        # Assert
        assert test_image.image_id == expected

    @staticmethod
    @mark.basic
    def test_image_id_cannot_be_reassigned(image: WhiteImage) -> None:
        """
        Once image_id is assigned, it cannot be reassigned
        :param image: test image
        """
        # Arrange
        test_image = Image(file_name=image.file_name, location=image.loc, external_id=image.external_id)

        # Act and assert
        with raises(ValueError):
            test_image.image_id = "ec2aac06-c261-3669-b8bd-4486a54ce740"

    @staticmethod
    @mark.basic
    def test_image_stores_correct_image_and_meta(image: WhiteImage) -> None:
        """
        Image stores image given as np.array correctly. It returns height as width as expected
        :param image: Test image
        """
        # Arrange
        test_image = Image(location=image.loc, file_name=image.file_name)

        # Act
        test_image.image = image.get_image_as_np_array()

        # Assert
        assert_array_equal(test_image.get_image().to_np_array(), image.get_image_as_np_array())
        assert test_image.height == image.get_bounding_box().height
        assert test_image.width == image.get_bounding_box().width

    @staticmethod
    @mark.basic
    def test_image_has_its_own_embedding_entry(image: WhiteImage) -> None:
        """
        Image has embedding k/v entry image_id/self._bbox
        :param image: Test image
        """

        # Arrange
        test_image = Image(location=image.loc, file_name=image.file_name)

        # Act
        test_image.image = image.get_image_as_np_array()

        # Assert
        assert test_image.get_embedding(test_image.image_id) == test_image._bbox  # pylint: disable=W0212

    @staticmethod
    @mark.basic
    def test_image_stores_correct_image_from_pdf(pdf_page: TestPdfPage) -> None:
        """

        :param pdf_page: Test pdf_page, represented in bytes
        """

        # Arrange
        test_image = Image(location=pdf_page.loc, file_name=pdf_page.file_name)

        # Act
        test_image.image = pdf_page.get_image_as_pdf_bytes()  # type: ignore

        # Assert
        assert test_image.height == pdf_page.np_array_shape[0]
        assert test_image.width == pdf_page.np_array_shape[1]

    @staticmethod
    @mark.basic
    def test_image_returns_image_representation(image: WhiteImage) -> None:
        """
        Image stores and returns image given as b64 string correctly. It returns height as width as expected.
        :param image: Test image
        """
        # Arrange
        test_image = Image(location=image.loc, file_name=image.file_name)
        test_image.image = image.get_image_as_b64_string()  # type: ignore

        # Assert
        assert_array_equal(test_image.get_image().to_np_array(), image.get_image_as_np_array())
        assert test_image.get_image().to_b64() == image.get_image_as_b64_string()
        assert test_image.height == image.get_bounding_box().height
        assert test_image.width == image.get_bounding_box().width

    @staticmethod
    @mark.basic
    def test_dump_cat_and_check_ann_id(image: WhiteImage) -> None:
        """
        Categories are dumped and annotation ids are correctly assigned.
        """

        # Arrange
        test_image = Image(location=image.loc, file_name=image.file_name)
        cat = ImageAnnotation(
            category_name="FOO",
            category_id="1",
            bounding_box=BoundingBox(ulx=1.0, uly=1.0, width=1.0, height=2.0, absolute_coords=True),
        )
        sub_cat_1 = CategoryAnnotation(category_name="BAK", category_id="2")

        # Act
        cat.dump_sub_category(get_type("BAK"), sub_cat_1, test_image.image_id)
        test_image.dump(cat)

        # Assert
        assert test_image.image_id == get_uuid(image.loc + image.file_name)
        assert (
            cat.annotation_id == "531191bc-3b48-3592-b4c3-70a0a5ac20aa"
        )  # get_uuid("FOOBounding Box ulx: 1.0 uly: 1.0 lrx: 2.0 lry: 3.090c05f37-a017-39cc-a178-b84f9d14ff48")
        assert sub_cat_1.annotation_id == "1377b99d-127d-366f-a7fc-ba25296fe4e5"

    @staticmethod
    @mark.basic
    def test_dump_same_annotation_not_possible(image: WhiteImage) -> None:
        """
        Same image annotations cannot be dumped twice
        """

        # Arrange
        test_image = Image(location=image.loc, file_name=image.file_name)
        cat = ImageAnnotation(
            category_name="FOO",
            category_id="1",
            bounding_box=BoundingBox(ulx=1.0, uly=1.0, width=1.0, height=2.0, absolute_coords=True),
        )

        # Act and Assert
        test_image.dump(cat)

        with raises(ValueError):
            test_image.dump(cat)

    @staticmethod
    @mark.basic
    def test_get_annotation(image: WhiteImage) -> None:
        """
        Annotations are returned by conditions.
        """

        # Arrange
        test_image = Image(location=image.loc, file_name=image.file_name)
        cat_1 = ImageAnnotation(
            category_name="FOO",
            category_id="1",
            bounding_box=BoundingBox(ulx=1.0, uly=1.0, width=1.0, height=2.0, absolute_coords=True),
        )
        cat_2 = ImageAnnotation(
            category_name="BAK",
            category_id="2",
            bounding_box=BoundingBox(ulx=2.0, uly=2.0, width=2.0, height=2.0, absolute_coords=True),
        )
        cat_3 = ImageAnnotation(
            category_name="BAK",
            category_id="1",
            bounding_box=BoundingBox(ulx=1.5, uly=2.4, width=3.0, height=9.0, absolute_coords=True),
        )

        test_image.dump(cat_1)
        test_image.dump(cat_2)
        test_image.dump(cat_3)

        # Act
        filtered_anns_1 = test_image.get_annotation(category_names="FOO")
        filtered_anns_1_ids = anns_to_ids(filtered_anns_1)
        filtered_anns_2 = test_image.get_annotation(category_names="BLA")
        filtered_anns_2_ids = anns_to_ids(filtered_anns_2)
        filtered_anns_3 = test_image.get_annotation(annotation_ids=[cat_1.annotation_id, cat_3.annotation_id])

        filtered_anns_3_ids = anns_to_ids(filtered_anns_3)
        filtered_anns_4 = test_image.get_annotation(
            annotation_ids=[cat_2.annotation_id, cat_3.annotation_id],
        )
        filtered_anns_4_ids = anns_to_ids(filtered_anns_4)
        filtered_anns_5 = test_image.get_annotation_iter()
        filtered_anns_5_ids = anns_to_ids(filtered_anns_5)

        # Assert
        assert set(filtered_anns_1_ids) == {cat_1.annotation_id}
        assert set(filtered_anns_2_ids) == set()
        assert set(filtered_anns_3_ids) == {cat_1.annotation_id, cat_3.annotation_id}
        assert set(filtered_anns_4_ids) == {cat_2.annotation_id, cat_3.annotation_id}
        assert set(filtered_anns_5_ids) == {cat_1.annotation_id, cat_2.annotation_id, cat_3.annotation_id}

    @staticmethod
    @mark.basic
    def test_image_ann_to_image(image: WhiteImage) -> None:
        """
        test  image_ann_to_image add attr: image to ImageAnnotation and generates Image instance correctly
        """

        # Arrange
        test_image = Image(location=image.loc, file_name=image.file_name)
        test_image.image = ones((24, 85, 3), dtype=float32)
        cat_1 = ImageAnnotation(
            category_name="FOO",
            bounding_box=BoundingBox(ulx=15.0, uly=20.0, width=10.0, height=8.0, absolute_coords=True),
        )
        test_image.dump(cat_1)
        annotation_id = cat_1.annotation_id

        # Act
        test_image.image_ann_to_image(annotation_id=annotation_id, crop_image=True)

        # Assert
        assert cat_1.image
        assert cat_1.image.height == 4
        assert cat_1.image.width == 10
        assert cat_1.image.get_embedding(test_image.image_id) == BoundingBox(
            ulx=15.0, uly=20.0, width=10.0, height=4.0, absolute_coords=True
        )
        assert cat_1.image.image.shape == (4, 10, 3)

    @staticmethod
    @mark.basic
    def test_load_image_from_dict() -> None:
        """
        test class  from_dict returns a image
        """
        df: DataFlow
        df = SerializerJsonlines.load(get_test_path() / "test_image.jsonl")
        df = MapData(df, lambda dp: Image.from_dict(**dp))
        image_list = collect_datapoint_from_dataflow(df)
        assert len(image_list) == 1

    @staticmethod
    @mark.basic
    def test_state_id_changes_when_annotations_added(image: WhiteImage) -> None:
        """
        state_id changes when annotations are added
        """

        # Arrange
        test_image = Image(location=image.loc, file_name=image.file_name)
        cat_1 = ImageAnnotation(
            category_name="FOO",
            category_id="1",
            bounding_box=BoundingBox(ulx=1.0, uly=1.0, width=1.0, height=2.0, absolute_coords=True),
        )
        cat_2 = ImageAnnotation(
            category_name="BAK",
            category_id="2",
            bounding_box=BoundingBox(ulx=2.0, uly=2.0, width=2.0, height=2.0, absolute_coords=True),
        )
        cat_3 = ImageAnnotation(
            category_name="BAK",
            category_id="1",
            bounding_box=BoundingBox(ulx=1.5, uly=2.4, width=3.0, height=9.0, absolute_coords=True),
        )

        test_image.dump(cat_1)
        test_image.dump(cat_2)

        # Assert
        assert test_image.state_id == "ec2d4ac5-a4dc-351f-b869-3ec6334d9906"

        # Act
        test_image.dump(cat_3)
        assert test_image.state_id == "ceb9021b-c96d-36e1-9cd5-465b48cae58b"



# -*- coding: utf-8 -*-
# File: test_eval.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module eval.eval
"""
from typing import List
from unittest.mock import MagicMock, patch

from pytest import fixture, mark

from deepdoctection.dataflow import DataFromList
from deepdoctection.datapoint import Image
from deepdoctection.datasets import DatasetCategories
from deepdoctection.eval import CocoMetric, Evaluator
from deepdoctection.extern.base import DetectionResult
from deepdoctection.pipe.layout import ImageLayoutService
from deepdoctection.utils import DatasetType, tensorpack_available

from ..test_utils import set_num_gpu_to_one

if tensorpack_available():
    from deepdoctection.extern.tpdetect import TPFrcnnDetector


class TestEvaluator:
    """
    Test Evaluator processes correctly
    """

    @fixture
    @patch("deepdoctection.extern.tp.tpcompat.get_num_gpu", MagicMock(side_effect=set_num_gpu_to_one))
    def setup_method(
        self,
        path_to_tp_frcnn_yaml: str,
        image_with_anns: Image,
        categories: DatasetCategories,
        detection_results: List[DetectionResult],
    ) -> None:
        """
        setup the necessary requirements
        """

        self._dataset = MagicMock()
        self._dataset.dataflow = MagicMock()
        self._dataset.dataset_info = MagicMock()
        self._dataset.dataflow.build = MagicMock(return_value=DataFromList([image_with_anns]))
        self._dataset.dataflow.categories = categories
        self._dataset.dataset_info.type = DatasetType.object_detection

        self._layout_detector = TPFrcnnDetector(
            path_yaml=path_to_tp_frcnn_yaml,
            path_weights="",
            categories=categories.get_categories(),
        )
        self._pipe_component = ImageLayoutService(self._layout_detector)
        self._pipe_component.predictor.predict = MagicMock(return_value=detection_results)  # type: ignore
        self._metric = CocoMetric

        self.evaluator = Evaluator(self._dataset, self._pipe_component, self._metric, 1)

    @mark.requires_tf
    @mark.full
    def test_evaluator_runs_and_returns_distance(self, setup_method) -> None:  #  type: ignore  # pylint: disable=W0613
        """
        Testing evaluator runs and returns metric distance
        """

        # Act
        out = self.evaluator.run()

        # Assert
        assert len(out) == 12



# -*- coding: utf-8 -*-
# File: pubtabnet.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for Pubtabnet dataset. Place the dataset as follows

    pubtabnet
    ├── test
    │ ├── PMC1.png
    ├── train
    │ ├── PMC2.png
    ├── val
    │ ├── PMC3.png
    ├── PubTabNet_2.0.0.jsonl
"""

from typing import Dict, List, Mapping, Union

from ...dataflow import DataFlow, MapData
from ...dataflow.custom_serialize import SerializerJsonlines
from ...datasets.info import DatasetInfo
from ...mapper.cats import cat_to_sub_cat, filter_cat
from ...mapper.pubstruct import pub_to_image
from ...utils.detection_types import JsonDict
from ...utils.logger import logger
from ...utils.settings import CellType, DatasetType, LayoutType, ObjectTypes, TableType, WordType
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories
from ..registry import dataset_registry

_NAME = "pubtabnet"
_DESCRIPTION = (
    "PubTabNet is a large dataset for image-based table recognition, containing 568k+ images of \n"
    "tabular data annotated with the corresponding HTML representation of the tables. The table images \n"
    " are extracted from the scientific publications included in the PubMed Central Open Access Subset \n"
    " (commercial use collection). Table regions are identified by matching the PDF format and \n"
    " the XML format of the articles in the PubMed Central Open Access Subset. More details are \n"
    " available in our paper 'Image-based table recognition: data, model, and evaluation'. \n"
    "Pubtabnet can be used for training cell detection models as well as for semantic table \n"
    "understanding algorithms. For detection it has cell bounding box annotations as \n"
    "well as precisely described table semantics like row - and column numbers and row and col spans. \n"
    "Moreover, every cell can be classified as header or non-header cell. The dataflow builder can also \n"
    "return captions of bounding boxes of rows and columns. Moreover, various filter conditions on \n"
    "the table structure are available: maximum cell numbers, maximal row and column numbers and their \n"
    "minimum equivalents can be used as filter condition"
)
_LICENSE = (
    "The annotations in this dataset belong to IBM and are licensed under a Community Data License Agreement \n"
    " – Permissive – Version 1.0 License. IBM does not own the copyright of the images. \n"
    " Use of the images must abide by the PMC Open Access Subset Terms of Use."
)
_URL = (
    "https://dax-cdn.cdn.appdomain.cloud/dax-pubtabnet/2.0.0/"
    "pubtabnet.tar.gz?_ga=2.267291150.146828643.1629125962-1173244232.1625045842"
)
_SPLITS: Mapping[str, str] = {"train": "train", "val": "val", "test": "test"}
_TYPE = DatasetType.object_detection
_LOCATION = "pubtabnet"
_ANNOTATION_FILES: Mapping[str, str] = {"all": "PubTabNet_2.0.0.jsonl"}

_INIT_CATEGORIES = [LayoutType.cell, TableType.item, LayoutType.table, LayoutType.word]
_SUB_CATEGORIES: Dict[ObjectTypes, Dict[ObjectTypes, List[ObjectTypes]]]
_SUB_CATEGORIES = {
    TableType.item: {TableType.item: [LayoutType.row, LayoutType.column]},
    LayoutType.cell: {
        CellType.header: [CellType.header, CellType.body],
        CellType.row_number: [],
        CellType.column_number: [],
        CellType.row_span: [],
        CellType.column_span: [],
        CellType.spanning: [CellType.spanning],
    },
    CellType.header: {
        CellType.row_number: [],
        CellType.column_number: [],
        CellType.row_span: [],
        CellType.column_span: [],
        CellType.spanning: [CellType.spanning],
    },
    CellType.body: {
        CellType.row_number: [],
        CellType.column_number: [],
        CellType.row_span: [],
        CellType.column_span: [],
        CellType.spanning: [CellType.spanning],
    },
    LayoutType.table: {TableType.html: [TableType.html]},
    LayoutType.word: {WordType.characters: [WordType.characters]},
}


@dataset_registry.register("pubtabnet")
class Pubtabnet(_BuiltInDataset):
    """
    Pubtabnet
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES, init_sub_categories=_SUB_CATEGORIES)

    def _builder(self) -> "PubtabnetBuilder":
        return PubtabnetBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class PubtabnetBuilder(DataFlowBaseBuilder):
    """
    Pubtabnet dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. Can be `train`, `val` or `test`. Default: `val`

        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`

        `load_image:` Will load the image for each datapoint.  Default: `False`

        `rows_and_cols:` Will add a 'item' image annotations that either represent a row or a column of a table.
                         Note, that the type of the item (i.e. being a row or a column) can be inferred from the
                         sub category added. Note further, that 'ITEM' are not originally part of the annotations
                         and are inferred from cell positions and their associated table semantic. Default: `True`

        `fake_score:` Will add a fake score so that annotations look like predictions

        :return: dataflow
        """
        split = str(kwargs.get("split", "val"))
        if split == "val":
            logger.info("Loading annotations for 'val' split from Pubtabnet will take some time.")
        max_datapoints = kwargs.get("max_datapoints")
        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)
            max_datapoints = max_datapoints if split == "train" else 500777 + max_datapoints
        load_image = kwargs.get("load_image", False)
        rows_and_cols = kwargs.get("rows_and_cols", False)
        fake_score = kwargs.get("fake_score", False)
        dd_pipe_like = kwargs.get("dd_pipe_like", False)
        if dd_pipe_like:
            logger.info("When 'dd_pipe_like'=True will set 'load_image'=True")
            load_image = True

        # Load
        df: DataFlow
        path = self.get_workdir() / self.get_annotation_file("all")
        df = SerializerJsonlines.load(path, max_datapoints=max_datapoints)

        # Map
        def replace_filename(dp: JsonDict) -> JsonDict:
            dp["filename"] = self.get_workdir() / dp["split"] / dp["filename"]
            return dp

        df = MapData(df, replace_filename)
        df = MapData(df, lambda dp: dp if dp["split"] == split else None)
        pub_mapper = pub_to_image(
            self.categories.get_categories(name_as_key=True, init=True),
            load_image,
            fake_score=fake_score,
            rows_and_cols=rows_and_cols,
            dd_pipe_like=dd_pipe_like,
            is_fintabnet=False,
            pubtables_like=False,
        )

        df = MapData(df, pub_mapper)
        if self.categories.is_cat_to_sub_cat():
            df = MapData(
                df,
                cat_to_sub_cat(self.categories.get_categories(name_as_key=True), self.categories.cat_to_sub_cat),
            )

        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_cat(  # pylint: disable=E1120
                    self.categories.get_categories(as_dict=False, filtered=True),
                    self.categories.get_categories(as_dict=False, filtered=False),
                ),
            )
        return df



# -*- coding: utf-8 -*-
# File: pdftext.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
PDFPlumber text extraction engine
"""

from typing import Dict, List, Tuple

from ..utils.context import save_tmp_file
from ..utils.detection_types import Requirement
from ..utils.file_utils import get_pdfplumber_requirement, pdfplumber_available
from ..utils.settings import LayoutType, ObjectTypes
from .base import DetectionResult, PdfMiner

if pdfplumber_available():
    from pdfplumber.pdf import PDF


def _to_detect_result(word: Dict[str, str]) -> DetectionResult:
    return DetectionResult(
        box=[float(word["x0"]), float(word["top"]), float(word["x1"]), float(word["bottom"])],
        class_id=1,
        text=word["text"],
        class_name=LayoutType.word,
    )


class PdfPlumberTextDetector(PdfMiner):
    """
    Text miner based on the pdfminer.six engine. To convert pdfminers result, especially group character to get word
    level results we use pdfplumber.

        pdf_plumber = PdfPlumberTextDetector()
        df = SerializerPdfDoc.load("path/to/document.pdf")

        for dp in df:
            detection_results = pdf_plumber.predict(dp["pdf_bytes"])

    To use it in a more integrated way:

        pdf_plumber = PdfPlumberTextDetector()
        text_extract = TextExtractionService(pdf_plumber)

        pipe = DoctectionPipe([text_extract])

        df = pipe.analyze(path="path/to/document.pdf")
        for dp in df:
            ...

    """

    def __init__(self) -> None:
        self.name = "pdfplumber"
        self.categories = {"1": LayoutType.word}

    def predict(self, pdf_bytes: bytes) -> List[DetectionResult]:
        """
        Call pdfminer.six and returns detected text as detection results

        :param pdf_bytes: bytes of a single pdf page
        :return: A list of DetectionResult
        """

        with save_tmp_file(pdf_bytes, "pdf_") as (tmp_name, _):
            with open(tmp_name, "rb") as fin:
                _pdf = PDF(fin)
                self._page = _pdf.pages[0]
                self._pdf_bytes = pdf_bytes
                words = self._page.extract_words()
        detect_results = list(map(_to_detect_result, words))
        return detect_results

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        return [get_pdfplumber_requirement()]

    def get_width_height(self, pdf_bytes: bytes) -> Tuple[float, float]:
        """
        Get the width and height of the full page
        :param pdf_bytes: pdf_bytes generating the pdf
        :return: width and height
        """

        if self._pdf_bytes == pdf_bytes:
            return self._page.bbox[2], self._page.bbox[3]
        # if the pdf bytes is not equal to the cached pdf, will recalculate values
        with save_tmp_file(pdf_bytes, "pdf_") as (tmp_name, _):
            with open(tmp_name, "rb") as fin:
                _pdf = PDF(fin)
                self._page = _pdf.pages[0]
                self._pdf_bytes = pdf_bytes
        return self._page.bbox[2], self._page.bbox[3]

    def possible_categories(self) -> List[ObjectTypes]:
        return [LayoutType.word]



# -*- coding: utf-8 -*-
# File: test_parallel_map.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module dataflow.parallel_map
"""
from typing import no_type_check

import numpy as np
from pytest import mark

from deepdoctection.dataflow import DataFlow, FakeData, MultiThreadMapData

from ..test_utils import collect_datapoint_from_dataflow


@no_type_check
def map_to_one(dp):
    """map to one"""
    return np.ones(dp[0].shape)


@mark.basic
def test_multithread_map_data_non_strict() -> None:
    """Test MultiThreadMapData non strict"""

    # Arrange
    df: DataFlow
    df = FakeData(shapes=[[4, 7, 3]], domain=(0, 1), size=10)

    # Act
    df = MultiThreadMapData(df, num_thread=4, map_func=map_to_one, buffer_size=10)
    output = collect_datapoint_from_dataflow(df, max_datapoints=20)

    # Assert
    assert len(output) == 20


@mark.basic
def test_multithread_map_data_strict() -> None:
    """Test MultiThreadMapData strict"""

    # Arrange
    df: DataFlow
    df = FakeData(shapes=[[4, 7, 3]], domain=(0, 1), size=10)

    # Act
    df = MultiThreadMapData(df, num_thread=4, map_func=map_to_one, buffer_size=11, strict=True)
    output = collect_datapoint_from_dataflow(df, max_datapoints=20)

    # Assert
    assert len(output) == 10



# -*- coding: utf-8 -*-
# File: model_fpn.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file is modified from
<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/modeling/model_fpn.py>
"""

import itertools

import numpy as np

# pylint: disable=import-error
import tensorflow as tf
from tensorpack import tfv1
from tensorpack.models import Conv2D, FixedUnPooling, MaxPooling, layer_register
from tensorpack.tfutils.argscope import argscope
from tensorpack.tfutils.scope_utils import under_name_scope
from tensorpack.tfutils.summary import add_moving_summary
from tensorpack.tfutils.tower import get_current_tower_context
from tensorpack.utils.argtools import memoized

from ..utils.box_ops import area as tf_area
from .backbone import GroupNorm
from .model_box import roi_align
from .model_rpn import generate_rpn_proposals, get_all_anchors, rpn_losses

# pylint: enable=import-error


@layer_register(log_shape=True)
def fpn_model(features, fpn_num_channels, fpn_norm):
    """
    Feature Pyramid Network model

    :param features: ResNet features c2-c5 [tf.Tensor]
    :param fpn_num_channels: FPN number of channels
    :param fpn_norm: FPN norm
    :return: FPN features p2-p6 [tf.Tensor]
    """
    assert len(features) == 4, features
    num_channel = fpn_num_channels

    use_gn = fpn_norm == "GN"

    def upsample2x(name, x):
        """
        Twofold up-sample

        :param name: name
        :param x: tf.Tensor
        :return: tf.Tensor
        """
        try:
            resize = tfv1.image.resize_images
            with tf.name_scope(name):
                shp2d = tf.shape(x)[2:]
                x = tf.transpose(x, [0, 2, 3, 1])
                x = resize(x, shp2d * 2, "nearest")
                x = tf.transpose(x, [0, 3, 1, 2])
                return x
        except AttributeError:
            return FixedUnPooling(name, x, 2, unpool_mat=np.ones((2, 2), dtype="float32"), data_format="channels_first")

    with argscope(
        Conv2D,
        data_format="channels_first",
        activation=tf.identity,
        use_bias=not use_gn,
        kernel_initializer=tfv1.variance_scaling_initializer(scale=1.0),
    ):
        lat_2345 = [Conv2D(f"lateral_1x1_c{i + 2}", c, num_channel, 1) for i, c in enumerate(features)]
        if use_gn:
            lat_2345 = [GroupNorm(f"gn_c{i + 2}", c) for i, c in enumerate(lat_2345)]
        lat_sum_5432 = []
        for idx, lat in enumerate(lat_2345[::-1]):
            if idx == 0:
                lat_sum_5432.append(lat)
            else:
                lat = lat + upsample2x(f"upsample_lat{6 - idx}", lat_sum_5432[-1])
                lat_sum_5432.append(lat)
        p2345 = [Conv2D(f"posthoc_3x3_p{i + 2}", c, num_channel, 3) for i, c in enumerate(lat_sum_5432[::-1])]
        if use_gn:
            p2345 = [GroupNorm(f"gn_p{i + 2}", c) for i, c in enumerate(p2345)]
        p6 = MaxPooling("maxpool_p6", p2345[-1], pool_size=1, strides=2, data_format="channels_first", padding="VALID")
        return p2345 + [p6]


@under_name_scope()
def fpn_map_rois_to_levels(boxes):
    """
    Assign boxes to level 2~5. Be careful that the returned tensor could be empty.

    :param boxes: (nx4)
    :return: 4 tensors for level 2-5. Each tensor is a vector of indices of boxes in its level. [tf.Tensor]
             4 tensors, the gathered boxes in each level. [tf.Tensor]
    """

    sqrtarea = tf.sqrt(tf_area(boxes))
    level = tf.cast(tf.floor(4 + tf.math.log(sqrtarea * (1.0 / 224) + 1e-6) * (1.0 / np.log(2))), tf.int32)

    # RoI levels range from 2~5 (not 6)
    level_ids = [
        tf.where(level <= 2),
        tf.where(tf.equal(level, 3)),  # == is not supported
        tf.where(tf.equal(level, 4)),
        tf.where(level >= 5),
    ]
    level_ids = [tf.reshape(x, [-1], name=f"roi_level{i + 2}_id") for i, x in enumerate(level_ids)]
    num_in_levels = [tf.size(x, name=f"num_roi_level{i + 2}") for i, x in enumerate(level_ids)]
    add_moving_summary(*num_in_levels)

    level_boxes = [tf.gather(boxes, ids) for ids in level_ids]
    return level_ids, level_boxes


@under_name_scope()
def multilevel_roi_align(features, rcnn_boxes, resolution, fpn_anchor_strides):
    """
    multilevel roi align


    :param resolution: output spatial resolution  (int)
    :param rcnn_boxes: nx4 boxes (tf.Tensor)
    :param features: 4 FPN feature level 2-5 ([tf.Tensor])
    :param fpn_anchor_strides: FPN anchor strides
    :return: NxC x res x res
    """

    assert len(features) == 4, features
    # Reassign rcnn_boxes to levels
    level_ids, level_boxes = fpn_map_rois_to_levels(rcnn_boxes)
    all_rois = []

    # Crop patches from corresponding levels
    for i, boxes, featuremap in zip(itertools.count(), level_boxes, features):
        with tf.name_scope(f"roi_level{i + 2}"):
            boxes_on_featuremap = boxes * (1.0 / fpn_anchor_strides[i])
            all_rois.append(roi_align(featuremap, boxes_on_featuremap, resolution))

    # this can fail if using TF<=1.8 with MKL build
    all_rois = tf.concat(all_rois, axis=0)  # NCHW
    # Unshuffle to the original order, to match the original samples
    level_id_perm = tf.concat(level_ids, axis=0)  # A permutation of 1~N
    level_id_invert_perm = tf.math.invert_permutation(level_id_perm)
    all_rois = tf.gather(all_rois, level_id_invert_perm, name="output")
    return all_rois


def multilevel_rpn_losses(
    multilevel_anchors, multilevel_label_logits, multilevel_box_logits, rpn_batch_per_im, fpn_anchor_strides
):
    """
    multilevel rpn losses

    :param multilevel_anchors: #lvl RPNAnchors
    :param multilevel_label_logits: #lvl tensors of shape HxWxA
    :param multilevel_box_logits: #lvl tensors of shape HxWxAx4
    :param rpn_batch_per_im: RPN batch per image
    :param fpn_anchor_strides: FPN anchor strides
    :return: label_loss, box_loss
    """

    num_lvl = len(fpn_anchor_strides)
    assert len(multilevel_anchors) == num_lvl
    assert len(multilevel_label_logits) == num_lvl
    assert len(multilevel_box_logits) == num_lvl

    losses = []
    with tf.name_scope("rpn_losses"):
        for lvl in range(num_lvl):
            anchors = multilevel_anchors[lvl]
            label_loss, box_loss = rpn_losses(
                anchors.gt_labels,
                anchors.encoded_gt_boxes(),
                multilevel_label_logits[lvl],
                multilevel_box_logits[lvl],
                rpn_batch_per_im,
                name_scope=f"level{lvl + 2}",
            )
            losses.extend([label_loss, box_loss])

        total_label_loss = tf.add_n(losses[::2], name="label_loss")
        total_box_loss = tf.add_n(losses[1::2], name="box_loss")
        add_moving_summary(total_label_loss, total_box_loss)
    return [total_label_loss, total_box_loss]


@under_name_scope()
def generate_fpn_proposals(
    multilevel_pred_boxes,
    multilevel_label_logits,
    image_shape2d,
    fpn_anchor_strides,
    fpn_proposal_mode,
    rpn_train_per_level_nms_topk,
    rpn_per_level_nms_topk,
    rpn_min_size,
    rpn_proposal_nms_thresh,
    rpn_train_pre_nms_top_k,
    rpn_test_pre_nms_top_k,
    rpn_train_post_nms_top_k,
    rpn_test_post_nms_top_k,
):
    """
    generate fpn proposals

    :param multilevel_pred_boxes: #lvl HxWxAx4 boxes
    :param multilevel_label_logits: #lvl tensors of shape HxWxA
    :param image_shape2d: image shape 2d
    :param rpn_test_post_nms_top_k: RPN inference post NMS top k
    :param rpn_train_post_nms_top_k: RPN train post NMS top k
    :param rpn_test_pre_nms_top_k: RPN test pre nms top k
    :param rpn_train_pre_nms_top_k: RPN train pre NMS top k
    :param rpn_proposal_nms_thresh: RPN proposals NMS thresh
    :param rpn_min_size: RPN min size
    :param rpn_per_level_nms_topk:  RPN NMS top k per level
    :param rpn_train_per_level_nms_topk:  RPN train per level NMS top k
    :param fpn_proposal_mode: FPN proposal mode
    :param fpn_anchor_strides: FPN anchor strides
    :return: boxes: kx4 float
             scores: k logits
    """

    num_lvl = len(fpn_anchor_strides)
    assert len(multilevel_pred_boxes) == num_lvl
    assert len(multilevel_label_logits) == num_lvl

    training = get_current_tower_context().is_training
    all_boxes = []
    all_scores = []
    if fpn_proposal_mode == "Level":
        fpn_nms_top_k = rpn_train_per_level_nms_topk if training else rpn_per_level_nms_topk
        for lvl in range(num_lvl):
            with tf.name_scope(f"Lvl{lvl + 2}"):
                pred_boxes_decoded = multilevel_pred_boxes[lvl]

                proposal_boxes, proposal_scores = generate_rpn_proposals(
                    tf.reshape(pred_boxes_decoded, [-1, 4]),
                    tf.reshape(multilevel_label_logits[lvl], [-1]),
                    image_shape2d,
                    rpn_min_size,
                    rpn_proposal_nms_thresh,
                    fpn_nms_top_k,
                )
                all_boxes.append(proposal_boxes)
                all_scores.append(proposal_scores)

        proposal_boxes = tf.concat(all_boxes, axis=0)  # nx4
        proposal_scores = tf.concat(all_scores, axis=0)  # n
        # Here we are different from Detectron.
        # Detectron picks top-k within the batch, rather than within an image, however we do not have a batch.
        proposal_topk = tf.minimum(tf.size(proposal_scores), fpn_nms_top_k)
        proposal_scores, topk_indices = tf.nn.top_k(proposal_scores, k=proposal_topk, sorted=False)
        proposal_boxes = tf.gather(proposal_boxes, topk_indices, name="all_proposals")
    else:
        for lvl in range(num_lvl):
            with tf.name_scope(f"Lvl{lvl + 2}"):
                pred_boxes_decoded = multilevel_pred_boxes[lvl]
                all_boxes.append(tf.reshape(pred_boxes_decoded, [-1, 4]))
                all_scores.append(tf.reshape(multilevel_label_logits[lvl], [-1]))
        all_boxes = tf.concat(all_boxes, axis=0)
        all_scores = tf.concat(all_scores, axis=0)
        proposal_boxes, proposal_scores = generate_rpn_proposals(
            all_boxes,
            all_scores,
            image_shape2d,
            rpn_min_size,
            rpn_proposal_nms_thresh,
            rpn_train_pre_nms_top_k if training else rpn_test_pre_nms_top_k,
            rpn_train_post_nms_top_k if training else rpn_test_post_nms_top_k,
        )

    tf.sigmoid(proposal_scores, name="probs")  # for visualization
    return tf.stop_gradient(proposal_boxes, name="boxes"), tf.stop_gradient(proposal_scores, name="scores")


@memoized
def get_all_anchors_fpn(*, strides, sizes, ratios, max_size):
    """
    get all anchors for fpn

    :return: each anchors is a SxSx NUM_ANCHOR_RATIOS x4 array. [anchors]
    """
    assert len(strides) == len(sizes)
    foas = []
    for stride, size in zip(strides, sizes):
        foa = get_all_anchors(stride=stride, sizes=(size,), ratios=ratios, max_size=max_size)
        foas.append(foa)
    return foas



# -*- coding: utf-8 -*-
# File: conftest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fixtures
"""

import itertools
from typing import List

import numpy as np
from pytest import fixture

from deepdoctection.datapoint import BoundingBox, Image, ImageAnnotation
from deepdoctection.datasets import DatasetCategories
from deepdoctection.extern.base import DetectionResult
from deepdoctection.utils.settings import get_type


@fixture(name="datapoint_image")
def fixture_datapoint_image() -> Image:
    """
    fixture Image datapoint
    """
    img = Image(location="/test/to/path", file_name="test_name")
    img.image = np.ones([400, 600, 3], dtype=np.float32)
    return img


@fixture(name="annotations")
def fixture_annotation() -> List[ImageAnnotation]:
    """
    annotations
    """
    row_anns = [
        ImageAnnotation(
            bounding_box=BoundingBox(ulx=15.0, uly=100.0, lrx=60.0, lry=150.0, absolute_coords=True),
            category_name="row",
            category_id="1",
        ),
        ImageAnnotation(
            bounding_box=BoundingBox(ulx=15.0, uly=200.0, lrx=70.0, lry=240.0, absolute_coords=True),
            category_name="row",
            category_id="1",
        ),
    ]

    col_anns = [
        ImageAnnotation(
            bounding_box=BoundingBox(ulx=10.0, uly=50.0, lrx=20.0, lry=250.0, absolute_coords=True),
            category_name="column",
            category_id="2",
        ),
        ImageAnnotation(
            bounding_box=BoundingBox(ulx=40.0, uly=20.0, lrx=50.0, lry=240.0, absolute_coords=True),
            category_name="column",
            category_id="2",
        ),
    ]
    return list(itertools.chain(row_anns, col_anns))


@fixture(name="image_with_anns")
def fixture_image_with_anns(datapoint_image: Image, annotations: List[ImageAnnotation]) -> Image:
    """
    image with annotations
    """
    for ann in annotations:
        datapoint_image.dump(ann)

    return datapoint_image


@fixture(name="categories")
def fixture_categories() -> DatasetCategories:
    """
    categories
    """
    return DatasetCategories(init_categories=[get_type("row"), get_type("column")])


@fixture(name="detection_results")
def fixture_detection_results() -> List[DetectionResult]:
    """
    detection results
    """
    detect_results_list = [
        DetectionResult(box=[15.0, 100.0, 60.0, 150.0], score=0.9, class_id=1, class_name=get_type("row")),
        DetectionResult(box=[15.0, 200.0, 70.0, 240.0], score=0.8, class_id=1, class_name=get_type("row")),
        DetectionResult(box=[10.0, 50.0, 20.0, 250.0], score=0.7, class_id=2, class_name=get_type("column")),
    ]

    return detect_results_list



# -*- coding: utf-8 -*-
# File: xxx.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for methods that might be helpful for testing
"""

import os
from pathlib import Path
from typing import Any, Iterable, Iterator, List, Optional, Union

from deepdoctection.dataflow import DataFlow
from deepdoctection.datapoint import Annotation


def get_test_path() -> Path:
    """
    get path to test objects
    """
    return Path(os.path.split(__file__)[0]) / "test_objects"


def get_integration_test_path() -> Path:
    """
    fixture integration test path
    """
    return get_test_path() / "sample_2"


def collect_datapoint_from_dataflow(
    df: Union[DataFlow, Iterator[Any]], max_datapoints: Optional[int] = None
) -> List[Any]:
    """
    Calls the reset_state method of a dataflow and collects all datapoints to an output list
    :param df: A Dataflow
    :param max_datapoints: The maximum number of datapoints to yield from
    :return: A list of datapoints of df
    """

    output: List[Any] = []
    if hasattr(df, "reset_state"):
        df.reset_state()  # type: ignore

    for idx, dp in enumerate(df):
        if max_datapoints is not None:
            if idx >= max_datapoints:
                break

        output.append(dp)

    return output


def anns_to_ids(annotations: Union[Iterable[Annotation], List[Annotation]]) -> List[Optional[str]]:
    """
    For a list of annotations return the list of annotation ids.
    :param annotations: A list of Annotations
    :return: A list of corresponding annotation ids
    """

    return [ann.annotation_id for ann in annotations]


def set_num_gpu_to_one() -> int:
    """
    set gpu number to one
    """
    return 1



# -*- coding: utf-8 -*-
# File: identifier.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Methods for generating and checking uuids
"""
import hashlib
import uuid

from .detection_types import Pathlike

__all__ = ["is_uuid_like", "get_uuid_from_str", "get_uuid"]


def is_uuid_like(input_id: str) -> bool:
    """
    Check, if input string has uuid3 string representation format.

    **Example:**

            "886313e1-3b8a-5372-9b90-0c9aee199e5d"

    :param input_id: An input string
    :return: A boolean output
    """
    try:
        uuid.UUID(str(input_id))
        return True
    except ValueError:
        return False


def get_uuid_from_str(input_id: str) -> str:
    """
    Returns an uuid3 string representation generated from an input string.

    :param input_id:
    :return: uuid3 string representation
    """
    return str(uuid.uuid3(uuid.NAMESPACE_DNS, input_id))


def get_uuid(*inputs: str) -> str:
    """
    Set an uuid generated by the concatenation of string inputs.

    :param inputs: string inputs
    :return: uuid3 string representation
    """
    str_input = "".join(inputs)
    return get_uuid_from_str(str_input)


def get_md5_hash(path: Pathlike, buffer_size: int = 65536) -> str:
    """
    Calculate a md5 hash for a given file

    :param path: path to a file
    :param buffer_size: Will calculate the hash in chunks
    :return: md5 string
    """

    hash_md5 = hashlib.md5()

    with open(path, "rb") as file:
        for chunk in iter(lambda: file.read(buffer_size), b""):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()



# -*- coding: utf-8 -*-
# File: fintabnet.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for Fintabnet dataset. Place the dataset as follows

    fintabnet
    ├── pdf
    │ ├── A
    │ │├── 2003
    │ │ ...
    ├── FinTabNet_1.0.0_cell_test.jsonl
    ├── FinTabNet_1.0.0_cell_train.jsonl
    ├── FinTabNet_1.0.0_cell_val.jsonl
    ├── FinTabNet_1.0.0_table_test.jsonl
    ├── FinTabNet_1.0.0_table_train.jsonl
    ├── FinTabNet_1.0.0_table_val.jsonl
"""

from pathlib import Path
from typing import List, Mapping, Sequence, Union

from ...dataflow import DataFlow, MapData, MultiProcessMapData
from ...dataflow.common import FlattenData
from ...dataflow.custom_serialize import SerializerJsonlines
from ...datapoint.image import Image
from ...mapper.cats import cat_to_sub_cat, filter_cat
from ...mapper.maputils import curry
from ...mapper.misc import image_ann_to_image, maybe_ann_to_sub_image
from ...mapper.pubstruct import pub_to_image
from ...utils.detection_types import JsonDict
from ...utils.file_utils import set_mp_spawn
from ...utils.logger import logger
from ...utils.settings import CellType, DatasetType, LayoutType, ObjectTypes, TableType
from ...utils.utils import to_bool
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories, DatasetInfo
from ..registry import dataset_registry

_NAME = "fintabnet"
_DESCRIPTION = (
    "FinTabNet dataset contains complex tables from the annual reports of S&P 500 companies with detailed \n"
    " table structure annotations to help train and test structure recognition. \n"
    "To generate the cell structure labels, one uses token matching between the PDF and HTML version \n"
    " of each article from public records and filings. Financial tables often have diverse styles when \n"
    "compared to ones in scientific and government documents, with fewer graphical lines and larger \n"
    " gaps within each table and more colour variations. Fintabnet can be used for training cell \n"
    "detection models as well as for semantic table understanding algorithms. \n"
    "For detection it has cell bounding box annotations as well as precisely described table semantics \n"
    "like row - and column numbers and row and col spans. The dataflow builder can also \n"
    "return captions of bounding boxes of rows and columns. Moreover, various filter conditions on \n"
    "the table structure are available: maximum cell numbers, maximal row and column numbers and their \n"
    "minimum equivalents can be used as filter condition. Header information of cells are not available. \n"
    "As work around you can artificially add header sub-category to every first row cell. \n"
    "All later row cells will receive a no header  sub-category. Note, that this assumption \n"
    "will generate noise."
)
_LICENSE = (
    "Community Data License Agreement – Permissive – Version 1.0  ---- \n"
    "This is the Community Data License Agreement – Permissive, Version 1.0 (“Agreement”).  \n"
    "Data is provided to You under this Agreement by each of the Data Providers.  Your exercise of any of \n"
    " the rights and permissions granted below constitutes Your acceptance and agreement to be bound by \n"
    " the terms and conditions of this Agreement."
)
_URL = (
    "https://dax-cdn.cdn.appdomain.cloud/dax-fintabnet/1.0.0/"
    "fintabnet.tar.gz?_ga=2.17492593.994196051.1634564576-1173244232.1625045842"
)
_SPLITS: Mapping[str, str] = {"train": "train", "val": "val", "test": "test"}
_TYPE = DatasetType.object_detection
_LOCATION = "fintabnet"
_ANNOTATION_FILES: Mapping[str, str] = {
    "train": "FinTabNet_1.0.0_table_train.jsonl",
    "test": "FinTabNet_1.0.0_table_test.jsonl",
    "val": "FinTabNet_1.0.0_table_val.jsonl",
}
_INIT_CATEGORIES = [LayoutType.table, LayoutType.cell, TableType.item]
_SUB_CATEGORIES: Mapping[ObjectTypes, Mapping[ObjectTypes, Sequence[ObjectTypes]]]
_SUB_CATEGORIES = {
    LayoutType.cell: {
        CellType.header: [CellType.header, CellType.body],
        CellType.row_number: [],
        CellType.column_number: [],
        CellType.row_span: [],
        CellType.column_span: [],
        CellType.spanning: [CellType.spanning, LayoutType.cell],
    },
    TableType.item: {TableType.item: [LayoutType.row, LayoutType.column]},
    CellType.header: {
        CellType.row_number: [],
        CellType.column_number: [],
        CellType.row_span: [],
        CellType.column_span: [],
        CellType.spanning: [CellType.spanning, LayoutType.cell],
    },
    CellType.body: {
        CellType.row_number: [],
        CellType.column_number: [],
        CellType.row_span: [],
        CellType.column_span: [],
        CellType.spanning: [CellType.spanning, LayoutType.cell],
    },
}


@dataset_registry.register("fintabnet")
class Fintabnet(_BuiltInDataset):
    """
    Fintabnet
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES, init_sub_categories=_SUB_CATEGORIES)

    def _builder(self) -> "FintabnetBuilder":
        return FintabnetBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class FintabnetBuilder(DataFlowBaseBuilder):
    """
    Fintabnet builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. Can be `train`,`val` or `test`. Default: `val`

        `build_mode:` Returns the full image or crops a table according to the table bounding box. Pass `table`
                           if you only want the cropped table. Default: ""

        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`

        `rows_and_cols:` Will add a 'item' image annotations that either represent a row or a column of a table.
                              Note, that the type of the item (i.e. being a row or a column) can be inferred from the
                              sub category added. Note further, that "item" are not originally part of the annotations
                              and are inferred from cell positions and their associated table semantic. Default: `True`
        `load_image:` Will load the image for each datapoint.  Default: `False`

        `use_multi_proc:` As the original files are stored as pdf conversion into a numpy array is time-consuming.
                        When setting use_multi_proc to True is will use several processes depending on the number
                        of CPUs available.

        `use_multi_proc_strict:` Will use strict mode in multiprocessing.

        `fake_score:` Will add a fake score so that annotations look like predictions

        :return: dataflow
        """

        split = str(kwargs.get("split", "val"))
        max_datapoints = kwargs.get("max_datapoints")
        rows_and_cols = kwargs.get("rows_and_cols", True)
        load_image = kwargs.get("load_image", False)
        use_multi_proc = to_bool(kwargs.get("use_multi_proc", True))
        use_multi_proc_strict = to_bool(kwargs.get("use_multi_proc_strict", False))
        fake_score = kwargs.get("fake_score", False)
        build_mode = kwargs.get("build_mode")
        pubtables_like = kwargs.get("pubtables_like", False)

        if build_mode and not load_image:
            logger.info("When 'build_mode' is set to True will reset 'load_image' to True")
            load_image = True

        if use_multi_proc or use_multi_proc_strict:
            set_mp_spawn()

        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)
        if kwargs.get("build_mode", "") != "table":
            logger.info("Logic will display only only table per page, even if there are more!!")

        # Load
        df: DataFlow
        path = self.get_workdir() / self.get_annotation_file(split)
        df = SerializerJsonlines.load(path, max_datapoints=max_datapoints)

        # Map
        @curry
        def _map_filename(dp: JsonDict, workdir: Path) -> JsonDict:
            dp["filename"] = workdir / "pdf" / dp["filename"]
            return dp

        map_filename = _map_filename(self.get_workdir())  # pylint: disable=E1120  # 259
        df = MapData(df, map_filename)

        buffer_size = 200 if max_datapoints is None else min(max_datapoints, 200) - 1

        pub_mapper = pub_to_image(
            self.categories.get_categories(name_as_key=True, init=True),
            load_image,
            fake_score=fake_score,
            rows_and_cols=rows_and_cols,
            dd_pipe_like=False,
            is_fintabnet=True,
            pubtables_like=pubtables_like,
        )
        if use_multi_proc:
            df = MultiProcessMapData(
                df,
                num_proc=1 if buffer_size < 3 else 4,
                map_func=pub_mapper,
                strict=use_multi_proc_strict,
                buffer_size=buffer_size,
            )
        else:
            df = MapData(df, pub_mapper)

        if build_mode == "table":

            @curry
            def _crop_and_add_image(dp: Image, category_names: List[str]) -> Image:
                return image_ann_to_image(dp, category_names=category_names)

            df = MapData(
                df,
                _crop_and_add_image(  # pylint: disable=E1120
                    category_names=[
                        LayoutType.table,
                        LayoutType.cell,
                        CellType.header,
                        CellType.body,
                        TableType.item,
                        LayoutType.row,
                        LayoutType.column,
                    ]
                ),
            )
            df = MapData(
                df,
                maybe_ann_to_sub_image(  # pylint: disable=E1120  # 259
                    category_names_sub_image=LayoutType.table,
                    category_names=[
                        LayoutType.cell,
                        CellType.header,
                        CellType.body,
                        TableType.item,
                        LayoutType.row,
                        LayoutType.column,
                    ],
                    add_summary=True,
                ),
            )
            df = MapData(df, lambda dp: [ann.image for ann in dp.get_annotation_iter(category_names=LayoutType.table)])
            df = FlattenData(df)
            df = MapData(df, lambda dp: dp[0])

        if self.categories.is_cat_to_sub_cat():
            df = MapData(
                df,
                cat_to_sub_cat(self.categories.get_categories(name_as_key=True), self.categories.cat_to_sub_cat),
            )

        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_cat(  # pylint: disable=E1120
                    self.categories.get_categories(as_dict=False, filtered=True),
                    self.categories.get_categories(as_dict=False, filtered=False),
                ),
            )
        return df



# -*- coding: utf-8 -*-
# File: test_refine.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module pipe.refine
"""

from typing import List, Set, Tuple

from pytest import mark

from deepdoctection.datapoint import ContainerAnnotation, Image
from deepdoctection.pipe.refine import (
    TableSegmentationRefinementService,
    _html_table,
    connected_component_tiles,
    generate_rectangle_tiling,
    rectangle_cells,
)

# from deepdoctection.utils.settings import names
from deepdoctection.utils.settings import CellType, LayoutType, TableType


@mark.basic
@mark.parametrize(
    "tiles_to_cells,expected_rectangle_cells_list",
    [
        (
            [
                ((1, 1), "a"),
                ((1, 2), "b"),
                ((1, 3), "b"),
                ((2, 1), "c"),
                ((2, 2), "d"),
                ((2, 3), "e"),
                ((3, 1), "f"),
                ((3, 2), "f"),
                ((3, 3), "f"),
            ],
            [{"d"}, {"e"}, {"f"}, {"b"}, {"a"}, {"c"}],
        ),
        (
            [
                ((1, 1), "a"),
                ((1, 2), "b"),
                ((1, 3), "b"),
                ((2, 1), "c"),
                ((2, 2), "d"),
                ((2, 3), "e"),
                ((2, 1), "f"),
                ((2, 2), "f"),
                ((2, 3), "f"),
            ],
            [{"a"}, {"f", "c", "d", "e"}, {"b"}],
        ),
        (
            [
                ((1, 1), "a"),
                ((1, 2), "b"),
                ((1, 2), "c"),
                ((2, 1), "d"),
                ((2, 2), "e"),
                ((2, 2), "c"),
                ((2, 1), "f"),
                ((2, 2), "f"),
                ((2, 2), "f"),
            ],
            [{"a", "c", "d", "e", "b", "f"}],
        ),
    ],
)
def test_rectangle_cell_tiling(
    tiles_to_cells: List[Tuple[Tuple[int, int], str]], expected_rectangle_cells_list: List[Set[str]]
) -> None:
    """
    Test that from a list of cell tiling one correctly receives a list of sets of cells that must be merged
    in order to have only rectangular shaped cells
    """
    # Act
    connected_components, tile_to_cell_dict = connected_component_tiles(tiles_to_cells)
    rectangle_tiling = generate_rectangle_tiling(connected_components)
    rectangle_cells_list = rectangle_cells(rectangle_tiling, tile_to_cell_dict)

    # Assert
    for el in expected_rectangle_cells_list:
        assert el in rectangle_cells_list

    for el in rectangle_cells_list:
        assert el in expected_rectangle_cells_list


@mark.basic
@mark.parametrize(
    "table_list,cells_ann_list,number_of_rows,number_of_cols,expected_html_list",
    [
        (
            [
                (1, [(1, 1, 1, 1), (1, 2, 1, 2)]),
                (2, [(2, 1, 2, 1), (2, 2, 1, 1), (2, 3, 1, 1)]),
                (3, [(3, 2, 1, 1), (3, 3, 1, 1)]),
                (4, [(4, 1, 1, 1), (4, 2, 1, 2)]),
            ],
            [(1, ["a", "b"]), (2, ["c", "d", "e"]), (3, ["f", "g"]), (4, ["h", "i"])],
            4,
            3,
            [
                "<table>",
                "<tr>",
                "<td>",
                "a",
                "</td>",
                "<td colspan=2>",
                "b",
                "</td>",
                "</tr>",
                "<tr>",
                "<td rowspan=2>",
                "c",
                "</td>",
                "<td>",
                "d",
                "</td>",
                "<td>",
                "e",
                "</td>",
                "</tr>",
                "<tr>",
                "<td>",
                "f",
                "</td>",
                "<td>",
                "g",
                "</td>",
                "</tr>",
                "<tr>",
                "<td>",
                "h",
                "</td>",
                "<td colspan=2>",
                "i",
                "</td>",
                "</tr>",
                "</table>",
            ],
        ),
        (
            [(1, [(1, 1, 3, 1), (1, 2, 1, 2)]), (2, [(2, 2, 1, 1), (2, 3, 1, 1)]), (3, [(3, 2, 1, 2)])],
            [(1, ["a", "b"]), (2, ["c", "d"]), (3, ["e"])],
            3,
            3,
            [
                "<table>",
                "<tr>",
                "<td rowspan=3>",
                "a",
                "</td>",
                "<td colspan=2>",
                "b",
                "</td>",
                "</tr>",
                "<tr>",
                "<td>",
                "c",
                "</td>",
                "<td>",
                "d",
                "</td>",
                "</tr>",
                "<tr>",
                "<td colspan=2>",
                "e",
                "</td>",
                "</tr>",
                "</table>",
            ],
        ),
        (
            [(1, [(1, 2, 1, 1), (1, 3, 1, 2)]), (2, [(2, 3, 2, 1)]), (3, [(3, 1, 2, 2)]), (4, [(4, 4, 1, 1)])],
            [(1, ["a", "b"]), (2, ["c"]), (3, ["d"]), (4, ["e"])],
            4,
            4,
            [
                "<table>",
                "<tr>",
                "<td>",
                "</td>",
                "<td>",
                "a",
                "</td>",
                "<td colspan=2>",
                "b",
                "</td>",
                "</tr>",
                "<tr>",
                "<td>",
                "</td>",
                "<td>",
                "</td>",
                "<td rowspan=2>",
                "c",
                "</td>",
                "<td>",
                "</td>",
                "</tr>",
                "<tr>",
                "<td rowspan=2 colspan=2>",
                "d",
                "</td>",
                "<td>",
                "</td>",
                "</tr>",
                "<tr>",
                "<td>",
                "</td>",
                "<td>",
                "e",
                "</td>",
                "</tr>",
                "</table>",
            ],
        ),
    ],
)
def test_generate_html_string(
    table_list: List[Tuple[int, List[Tuple[int, int, int, int]]]],
    cells_ann_list: List[Tuple[int, List[str]]],
    number_of_rows: int,
    number_of_cols: int,
    expected_html_list: List[str],
) -> None:
    """
    test _html_table
    """
    # Act
    html_list = _html_table(table_list, cells_ann_list, number_of_rows, number_of_cols)

    # Assert
    assert html_list == expected_html_list


class TestTableSegmentationRefinementService:
    """
    Test TableSegmentationRefinementService (only integrated)
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self.table_segmentation_refinement_service = TableSegmentationRefinementService()

    @mark.basic
    def test_integration_pipeline_component(self, dp_image_fully_segmented_fully_tiled: Image) -> None:
        """
        Integration test for pipeline component
        """

        # Arrange
        dp = dp_image_fully_segmented_fully_tiled

        # Act
        dp = self.table_segmentation_refinement_service.pass_datapoint(dp)

        # Assert
        table = dp.get_annotation(category_names=LayoutType.table)[0]
        assert table.image is not None
        summary = table.image.summary
        summaries_table = [
            summary.get_sub_category(TableType.number_of_rows).category_id,
            summary.get_sub_category(TableType.number_of_columns).category_id,
            summary.get_sub_category(TableType.max_row_span).category_id,
            summary.get_sub_category(TableType.max_col_span).category_id,
        ]
        summary_html = table.get_sub_category(TableType.html)
        cells = dp.get_annotation(
            category_names=self.table_segmentation_refinement_service._cell_names  # pylint: disable=W0212
        )
        row_numbers = {cell.get_sub_category(CellType.row_number).category_id for cell in cells}
        col_numbers = {cell.get_sub_category(CellType.column_number).category_id for cell in cells}
        row_spans = {cell.get_sub_category(CellType.row_span).category_id for cell in cells}
        col_spans = {cell.get_sub_category(CellType.column_span).category_id for cell in cells}

        assert len(cells) == 4
        assert row_numbers == {"1", "2"}
        assert col_numbers == {"1", "2"}
        assert row_spans == {"1", "1", "1", "1"}
        assert col_spans == {"1", "1", "1", "1"}
        assert summaries_table == ["2", "2", "1", "1"]
        assert isinstance(summary_html, ContainerAnnotation)
        assert summary_html.value == [
            "<table>",
            "<tr>",
            "<td>",
            "2c1458bf-6b21-327a-9632-21373da468bb",
            "</td>",
            "<td>",
            "609243a2-0bd8-3002-939a-c8bcf7b2e3c8",
            "</td>",
            "</tr>",
            "<tr>",
            "<td>",
            "9b2b0bd1-3885-3aa6-8d45-ebebdd1abddd",
            "</td>",
            "<td>",
            "6e70cf3d-e260-3f73-81bf-77616b6d54b2",
            "</td>",
            "</tr>",
            "</table>",
        ]
